__all__ = [
    'proto_to_dict',
    'dataclass_to_dict',
]
from protoplasm.casting import dictators
from protoplasm.structs import *

from google.protobuf import json_format

import logging
log = logging.getLogger(__name__)


def proto_to_dict(proto: GeneratedProtocolMessageType) -> Dict:
    """Turns a Protobuf message object into a dict.
    """
    return json_format.MessageToDict(proto,
                                     preserving_proto_field_name=True,
                                     use_integers_for_enums=True)


def dataclass_to_dict(dc) -> Dict:
    """Turns a Dataclass object into a dict with the data values serialized so
    that they can be turned into Protobuf objects.

    This will also turn non-proto related dataclasses into dicts but those won't
     necessarily be usable to build Protobuf messages as this will simply use
     the asdict call in the standard dataclass library.

    :param dc: The dataclass to turn into a dict
    :return: A dictionary of Proto-ready values.
    """
    if not hasattr(dc, '__proto_cls__'):  # Plain dataclass
        return dataclasses.asdict(dc)

    d = {}
    for field in dataclasses.fields(dc):
        dicted = _dataclass_field_to_dict_field(field, dc)
        # if dicted:
        if dicted is not None:  # TODO(thordurm@ccpgames.com>) 2024-04-15: To include "default/empty" fields or not?
            d[_get_proto_field_name(field)] = dicted

    return d


def _get_proto_field_name(field: dataclasses.Field):
    return field.metadata.get('pb_name', field.name)


def _dataclass_field_to_dict_field(field: dataclasses.Field, dc):
    val = getattr(dc, field.name)

    if val is None:
        return None

    dictator_cls = field.metadata.get('dictator', dictators.BaseDictator)
    is_obj = field.metadata.get('is_obj', False)
    is_list = field.metadata.get('is_list', False)
    is_map = field.metadata.get('is_map', False)

    if is_obj:
        if is_map:
            return {k: dataclass_to_dict(v) for k, v in val.items()} if val else None
        elif is_list:
            return [dataclass_to_dict(v) for v in val] if val else None
        else:
            return dataclass_to_dict(val)
    else:
        if is_map:
            return {k: dictator_cls.to_dict_value(v, field, dc) for k, v in val.items()} if val else None
        elif is_list:
            return [dictator_cls.to_dict_value(v, field, dc) for v in val] if val else None
        else:
            return dictator_cls.to_dict_value(val, field, dc)



__all__ = [
    'BaseGrpcClientImplementation',
]
import grpc  # noqa

from protoplasm.structs import *
from ._remotewrap import *
from .dataclass_bases import *

import logging
log = logging.getLogger(__name__)


class BaseGrpcClientImplementation:
    def __init__(self, service_stub_cls: Type, grpc_host='localhost:50051',
                 credentials: Optional[Union[bool, grpc.ChannelCredentials]] = None,
                 options: Optional[Dict] = None, *args, **kwargs):
        """
        credentials:
        - None -> Automatic... will use `grpc.ssl_channel_credentials()` and secure_channel if port is 443
        - True -> Will use `grpc.ssl_channel_credentials()` and secure_channel
        - False -> Will use insecure_channel and no creds
        - grpc.ChannelCredentials -> Will use given creds and secure_channel
        """
        options = options or {}
        if 'grpc.enable_http_proxy' not in options:
            options['grpc.enable_http_proxy'] = 0
        option_tuple = tuple(options.items())

        # Backwards compatibility
        if kwargs:
            if not credentials and 'credential' in kwargs:
                credentials = kwargs.pop('credential')
                log.warning('Deprecated keyword "credential" used when initiating BaseGrpcClient.'
                            ' Please use "credentials" instead!')

        # This keeps happening...!
        if ':' not in grpc_host:
            if credentials:
                log.warning('No port used in host... adding 443 by default (cause this is a secure connection)!')
                grpc_host = f'{grpc_host}:443'
            else:
                log.warning('No port used in host... adding 50051 by default!')
                grpc_host = f'{grpc_host}:50051'

        if credentials is None:
            if grpc_host.endswith(':443'):
                credentials = grpc.ssl_channel_credentials()
        elif credentials is True:
            credentials = grpc.ssl_channel_credentials()

        if credentials:
            self.grpc_channel: grpc.Channel = grpc.secure_channel(grpc_host, credentials, options=option_tuple)
        else:
            self.grpc_channel: grpc.Channel = grpc.insecure_channel(grpc_host, options=option_tuple)

        self.stub = service_stub_cls(self.grpc_channel)

    def _forward_to_grpc(self, request_cls: Type[T_DCB], stub_method: Callable,
                         *args, **kwargs) -> Optional[Union[Any, Tuple[Any], Iterable[T_DCB]]]:
        return RemoteMethodWrapper(self, request_cls, stub_method, *args, **kwargs).call()



import unittest
import datetime
import base64
import protoplasm.casting
from protoplasm.casting import castutils
import os
import sys
import shutil
import time

import logging
log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

HERE = os.path.dirname(__file__)

PROTO_ROOT = os.path.join(HERE, 'res', 'proto')
BUILD_ROOT = os.path.join(HERE, 'res', 'build')


class CastutilsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Remove old stuff...
        build_package = os.path.join(BUILD_ROOT, 'sandbox')
        if os.path.exists(build_package):
            shutil.rmtree(build_package)
            time.sleep(0.1)

        from neobuilder.neobuilder import NeoBuilder

        # Build stuff...
        builder = NeoBuilder(package='sandbox',
                             protopath=PROTO_ROOT,
                             build_root=BUILD_ROOT)
        builder.build()

        # Add build root to path to access its modules
        sys.path.append(BUILD_ROOT)

    def test_kwdict(self):
        self.assertEqual({}, castutils.kwdict())
        self.assertEqual({'foo': 'bar'}, castutils.kwdict(foo='bar'))
        self.assertEqual({'foo': 'bar', 'bar': 'foo'}, castutils.kwdict(foo='bar', bar='foo'))
        self.assertEqual({'bar': 1, 'foo': 'two', 'oof': None}, castutils.kwdict(oof=None, foo='two', bar=1))

        self.assertNotEqual({'bar': 1, 'foo': 'two'}, castutils.kwdict(oof=None, foo='two', bar=1))

    def test_nested_kwdict(self):
        self.assertEqual({'foo': 'bar', 'bar': {'one': 1, 'two': 2}}, castutils.kwdict(foo='bar', bar=castutils.kwdict(one=1, two=2)))
        self.assertEqual({'foo': 'bar', 'bar': {'one': 1, 'two': 2}}, castutils.kwdict(foo='bar', bar__one=1, bar__two=2))
        self.assertEqual({'foo': 'bar', 'bar': {'one': 1, 'two': {'more': 2, 'less': 3}}}, castutils.kwdict(foo='bar', bar__one=1, bar__two__more=2, bar__two__less=3))

        self.assertNotEqual({'foo': 'bar', 'bar': {'one': 1, 'two': {'more': 2, 'less': 3}}}, castutils.kwdict(foo='bar', bar__one=1, bar__two__more=2, bar__two_less=3))

    def test_big_kwdict(self):
        expected = {'simple_field': 'Green',
                    'message_field': {
                        'foo': 'Blue',
                        'bar': 'Yellow'
                    },
                    'simple_list': ['Four', 'Seven', '99'],
                    'message_list': [{'foo': 'Foo in a list',
                                      'bar': 'Candybar'},
                                     {'foo': 'a Fool in a list',
                                      'bar': 'Tequila bar'}],
                    'simple_map': {'dora': 'Imamap!',
                                   'diego': 'Camera!'},
                    'message_map': {'mickey': {'foo': 'mouse'},
                                    'donald': {'foo': 'duck',
                                               'bar': 'trump'}}}

        self.assertEqual(expected,
                         castutils.kwdict(simple_map__diego='Camera!',
                                          simple_field='Green',
                                          message_field__foo='Blue',
                                          message_field__bar='Yellow',
                                          simple_list=['Four', 'Seven', '99'],
                                          message_list=[castutils.kwdict(foo='Foo in a list',
                                                                         bar='Candybar'),
                                                        castutils.kwdict(foo='a Fool in a list',
                                                                         bar='Tequila bar')],
                                          simple_map__dora='Imamap!',
                                          message_map__donald__foo='duck',
                                          message_map__donald__bar='trump',
                                          message_map__mickey__foo='mouse'))

    def test_mkproto(self):
        from sandbox.test import rainbow_pb2

        p1 = rainbow_pb2.SubMessage()
        p1.foo = 'Foo!'
        p1.bar = 'Bar!!!'

        self.assertEqual(p1, protoplasm.casting.mkproto(rainbow_pb2.SubMessage, foo='Foo!', bar='Bar!!!'))

        p2 = rainbow_pb2.SubMessage()
        p2.foo = 'Foo Two!'
        self.assertEqual(p2, protoplasm.casting.mkproto(rainbow_pb2.SubMessage, foo='Foo Two!'))

        p3 = rainbow_pb2.SubMessage()
        p3.foo = 'Not Foo!'
        self.assertNotEqual(p3, protoplasm.casting.mkproto(rainbow_pb2.SubMessage, foo='Foo You!'))

        p4 = rainbow_pb2.SubMessage()
        self.assertEqual(p4, protoplasm.casting.mkproto(rainbow_pb2.SubMessage))

    def test_nested_mkproto(self):
        from sandbox.test import rainbow_pb2

        p1 = rainbow_pb2.RainbowMessage()
        self.assertEqual(p1, protoplasm.casting.mkproto(rainbow_pb2.RainbowMessage))

        p2 = rainbow_pb2.RainbowMessage()
        p2.simple_field = 'Green'
        self.assertEqual(p2, protoplasm.casting.mkproto(rainbow_pb2.RainbowMessage, simple_field='Green'))
        self.assertNotEqual(p1, protoplasm.casting.mkproto(rainbow_pb2.RainbowMessage, simple_field='Green'))

        p3 = rainbow_pb2.RainbowMessage()
        p3.simple_field = 'Green'
        p3.message_field.foo = 'Blue'
        self.assertEqual(p3, protoplasm.casting.mkproto(rainbow_pb2.RainbowMessage,
                                                        simple_field='Green',
                                                        message_field__foo='Blue'))
        p4 = rainbow_pb2.RainbowMessage()
        p4.simple_field = 'Green'
        p4.message_field.foo = 'Blue'
        p4.message_field.bar = 'Yellow'
        self.assertEqual(p4, protoplasm.casting.mkproto(rainbow_pb2.RainbowMessage,
                                                        simple_field='Green',
                                                        message_field__foo='Blue',
                                                        message_field__bar='Yellow'))
        self.assertNotEqual(p3, p4)
        self.assertNotEqual(p4, protoplasm.casting.mkproto(rainbow_pb2.RainbowMessage,
                                                           simple_field='Green',
                                                           message_field__foo='Blue',
                                                           message_field__bar='Red'))

        p5 = rainbow_pb2.RainbowMessage()
        p5.simple_field = 'Green'
        p5.message_field.foo = 'Blue'
        p5.message_field.bar = 'Yellow'
        p5.simple_list.append('Four')
        p5.simple_list.append('Seven')
        p5.simple_list.append('99')
        self.assertEqual(p5, protoplasm.casting.mkproto(rainbow_pb2.RainbowMessage,
                                                        simple_field='Green',
                                                        message_field__foo='Blue',
                                                        message_field__bar='Yellow',
                                                        simple_list=['Four', 'Seven', '99']))

        p6 = rainbow_pb2.RainbowMessage()
        p6.simple_field = 'Green'
        p6.message_field.foo = 'Blue'
        p6.message_field.bar = 'Yellow'
        p6.simple_list.append('Four')
        p6.simple_list.append('Seven')
        p6.simple_list.append('98')
        self.assertNotEqual(p6, protoplasm.casting.mkproto(rainbow_pb2.RainbowMessage,
                                                           simple_field='Green',
                                                           message_field__foo='Blue',
                                                           message_field__bar='Yellow',
                                                           simple_list=['Four', 'Seven', '100']))

        p7 = rainbow_pb2.RainbowMessage()
        p7.simple_field = 'Green'
        p7.message_field.foo = 'Blue'
        p7.message_field.bar = 'Yellow'
        p7.simple_list.append('Four')
        p7.simple_list.append('Seven')
        p7.simple_list.append('99')
        m1 = p7.message_list.add()
        m1.foo = 'Foo in a list'
        m1.bar = 'Candybar'
        m2 = p7.message_list.add()
        m2.foo = 'a Fool in a list'
        m2.bar = 'Tequila bar'
        self.assertEqual(p7, protoplasm.casting.mkproto(rainbow_pb2.RainbowMessage,
                                                        simple_field='Green',
                                                        message_field__foo='Blue',
                                                        message_field__bar='Yellow',
                                                        simple_list=['Four', 'Seven', '99'],
                                                        message_list=[castutils.kwdict(foo='Foo in a list',
                                                                              bar='Candybar'),
                                                             castutils.kwdict(foo='a Fool in a list',
                                                                              bar='Tequila bar')]))

        self.assertNotEqual(p7, protoplasm.casting.mkproto(rainbow_pb2.RainbowMessage,
                                                           simple_field='Green',
                                                           message_field__foo='Blue',
                                                           message_field__bar='Yellow',
                                                           simple_list=['Four', 'Seven', '99'],
                                                           message_list=[castutils.kwdict(foo='a Fool in a list',
                                                                                 bar='Tequila bar')]))

        p8 = rainbow_pb2.RainbowMessage()
        p8.simple_field = 'Green'
        p8.message_field.foo = 'Blue'
        p8.message_field.bar = 'Yellow'
        p8.simple_list.append('Four')
        p8.simple_list.append('Seven')
        p8.simple_list.append('99')
        m3 = p8.message_list.add()
        m3.foo = 'Foo in a list'
        m3.bar = 'Candybar'
        m4 = p8.message_list.add()
        m4.foo = 'a Fool in a list'
        m4.bar = 'Tequila bar'
        p8.simple_map['dora'] = 'Imamap!'
        p8.simple_map['diego'] = 'Camera!'
        self.assertEqual(p8, protoplasm.casting.mkproto(rainbow_pb2.RainbowMessage,
                                                        simple_field='Green',
                                                        message_field__foo='Blue',
                                                        message_field__bar='Yellow',
                                                        simple_list=['Four', 'Seven', '99'],
                                                        message_list=[castutils.kwdict(foo='Foo in a list',
                                                                              bar='Candybar'),
                                                             castutils.kwdict(foo='a Fool in a list',
                                                                              bar='Tequila bar')],
                                                        simple_map__dora='Imamap!',
                                                        simple_map__diego='Camera!'))

        p9 = rainbow_pb2.RainbowMessage()
        p9.simple_field = 'Green'
        p9.message_field.foo = 'Blue'
        p9.message_field.bar = 'Yellow'
        p9.simple_list.append('Four')
        p9.simple_list.append('Seven')
        p9.simple_list.append('99')
        m5 = p9.message_list.add()
        m5.foo = 'Foo in a list'
        m5.bar = 'Candybar'
        m6 = p9.message_list.add()
        m6.foo = 'a Fool in a list'
        m6.bar = 'Tequila bar'
        p9.simple_map['dora'] = 'Imamap!'
        p9.simple_map['diego'] = 'Camera!'
        p9.message_map['donald'].foo = 'duck'
        p9.message_map['donald'].bar = 'trump'
        p9.message_map['mickey'].foo = 'mouse'
        self.assertEqual(p9, protoplasm.casting.mkproto(rainbow_pb2.RainbowMessage,
                                                        simple_field='Green',
                                                        message_field__foo='Blue',
                                                        message_field__bar='Yellow',
                                                        simple_list=['Four', 'Seven', '99'],
                                                        message_list=[castutils.kwdict(foo='Foo in a list',
                                                                              bar='Candybar'),
                                                             castutils.kwdict(foo='a Fool in a list',
                                                                              bar='Tequila bar')],
                                                        simple_map__dora='Imamap!',
                                                        simple_map__diego='Camera!',
                                                        message_map__donald__foo='duck',
                                                        message_map__donald__bar='trump',
                                                        message_map__mickey__foo='mouse'))

        p10 = rainbow_pb2.RainbowMessage()
        p10.simple_field = 'Green'
        p10.message_field.foo = 'Blue'
        p10.message_field.bar = 'Yellow'
        p10.simple_list.append('Four')
        p10.simple_list.append('Seven')
        p10.simple_list.append('99')
        m7 = p10.message_list.add()
        m7.foo = 'Foo in a list'
        m7.bar = 'Candybar'
        m8 = p10.message_list.add()
        m8.foo = 'a Fool in a list'
        m8.bar = 'Tequila bar'
        p10.simple_map['dora'] = 'Imamap!'
        p10.simple_map['diego'] = 'Camera!'
        p10.message_map['mickey'].foo = 'mouse'
        p10.message_map['donald'].foo = 'duck'
        p10.message_map['donald'].bar = 'trump'

        self.assertEqual(p10, protoplasm.casting.mkproto(rainbow_pb2.RainbowMessage,
                                                         simple_map__diego='Camera!',
                                                         simple_field='Green',
                                                         message_field__foo='Blue',
                                                         message_field__bar='Yellow',
                                                         simple_list=['Four', 'Seven', '99'],
                                                         message_list=[castutils.kwdict(foo='Foo in a list',
                                                                               bar='Candybar'),
                                                              castutils.kwdict(foo='a Fool in a list',
                                                                               bar='Tequila bar')],
                                                         simple_map__dora='Imamap!',
                                                         message_map__donald__foo='duck',
                                                         message_map__donald__bar='trump',
                                                         message_map__mickey__foo='mouse'))

    def test_mkdataclass(self):
        from sandbox.test import rainbow_dc

        dc1 = rainbow_dc.SubMessage()
        dc1.foo = 'Foo!'
        dc1.bar = 'Bar!!!'

        self.assertEqual(dc1, protoplasm.casting.mkdataclass(rainbow_dc.SubMessage, foo='Foo!', bar='Bar!!!'))

        dc2 = rainbow_dc.SubMessage()
        dc2.foo = 'Foo Two!'
        self.assertEqual(dc2, protoplasm.casting.mkdataclass(rainbow_dc.SubMessage, foo='Foo Two!'))

        dc3 = rainbow_dc.SubMessage()
        dc3.foo = 'Not Foo!'
        self.assertNotEqual(dc3, protoplasm.casting.mkdataclass(rainbow_dc.SubMessage, foo='Foo You!'))

        dc4 = rainbow_dc.SubMessage()
        self.assertEqual(dc4, protoplasm.casting.mkdataclass(rainbow_dc.SubMessage))

    def test_nested_mkdataclass(self):
        from sandbox.test import rainbow_dc

        dc1 = rainbow_dc.RainbowMessage()
        self.assertEqual(dc1, protoplasm.casting.mkdataclass(rainbow_dc.RainbowMessage))

        dc2 = rainbow_dc.RainbowMessage()
        dc2.simple_field = 'Green'
        self.assertEqual(dc2, protoplasm.casting.mkdataclass(rainbow_dc.RainbowMessage, simple_field='Green'))
        self.assertNotEqual(dc1,
                            protoplasm.casting.mkdataclass(rainbow_dc.RainbowMessage, simple_field='Green'))

        dc3 = rainbow_dc.RainbowMessage()
        dc3.simple_field = 'Green'
        dc3.message_field = rainbow_dc.SubMessage()
        dc3.message_field.foo = 'Blue'

        expected_dc3 = protoplasm.casting.mkdataclass(rainbow_dc.RainbowMessage,
                                                      simple_field='Green',
                                                      message_field__foo='Blue')

        self.assertEqual(dc3, expected_dc3)

        dc4 = rainbow_dc.RainbowMessage()
        dc4.simple_field = 'Green'
        dc4.message_field = rainbow_dc.SubMessage()
        dc4.message_field.foo = 'Blue'
        dc4.message_field.bar = 'Yellow'
        self.assertEqual(dc4, protoplasm.casting.mkdataclass(rainbow_dc.RainbowMessage,
                                                             simple_field='Green',
                                                             message_field__foo='Blue',
                                                             message_field__bar='Yellow'))
        self.assertNotEqual(dc3, dc4)
        self.assertNotEqual(dc4, protoplasm.casting.mkdataclass(rainbow_dc.RainbowMessage,
                                                                simple_field='Green',
                                                                message_field__foo='Blue',
                                                                message_field__bar='Red'))

        dc5 = rainbow_dc.RainbowMessage()
        dc5.simple_field = 'Green'
        dc5.message_field = rainbow_dc.SubMessage('Blue', 'Yellow')
        dc5.simple_list.append('Four')
        dc5.simple_list.append('Seven')
        dc5.simple_list.append('99')
        self.assertEqual(dc5, protoplasm.casting.mkdataclass(rainbow_dc.RainbowMessage,
                                                             simple_field='Green',
                                                             message_field__foo='Blue',
                                                             message_field__bar='Yellow',
                                                             simple_list=['Four', 'Seven', '99']))

        dc6 = rainbow_dc.RainbowMessage()
        dc6.simple_field = 'Green'
        dc6.message_field = rainbow_dc.SubMessage(foo='Blue', bar='Yellow')
        dc6.simple_list.append('Four')
        dc6.simple_list.append('Seven')
        dc6.simple_list.append('98')
        self.assertNotEqual(dc6, protoplasm.casting.mkdataclass(rainbow_dc.RainbowMessage,
                                                                simple_field='Green',
                                                                message_field__foo='Blue',
                                                                message_field__bar='Yellow',
                                                                simple_list=['Four', 'Seven', '100']))

        dc7 = rainbow_dc.RainbowMessage()
        dc7.simple_field = 'Green'
        dc7.message_field = rainbow_dc.SubMessage(foo='Blue', bar='Yellow')
        dc7.simple_list.append('Four')
        dc7.simple_list.append('Seven')
        dc7.simple_list.append('99')

        dc7.message_list.append(rainbow_dc.SubMessage(foo='Foo in a list', bar='Candybar'))
        dc7.message_list.append(rainbow_dc.SubMessage(foo='a Fool in a list', bar='Tequila bar'))

        self.assertEqual(dc7, protoplasm.casting.mkdataclass(rainbow_dc.RainbowMessage,
                                                             simple_field='Green',
                                                             message_field__foo='Blue',
                                                             message_field__bar='Yellow',
                                                             simple_list=['Four', 'Seven', '99'],
                                                             message_list=[castutils.kwdict(foo='Foo in a list',
                                                                                            bar='Candybar'),
                                                                           castutils.kwdict(foo='a Fool in a list',
                                                                                            bar='Tequila bar')]))

        self.assertNotEqual(dc7, protoplasm.casting.mkdataclass(rainbow_dc.RainbowMessage,
                                                                simple_field='Green',
                                                                message_field__foo='Blue',
                                                                message_field__bar='Yellow',
                                                                simple_list=['Four', 'Seven', '99'],
                                                                message_list=[castutils.kwdict(foo='a Fool in a list',
                                                                                               bar='Tequila bar')]))

        dc8 = rainbow_dc.RainbowMessage()
        dc8.simple_field = 'Green'
        dc8.message_field = rainbow_dc.SubMessage(foo='Blue', bar='Yellow')
        dc8.simple_list.append('Four')
        dc8.simple_list.append('Seven')
        dc8.simple_list.append('99')

        dc8.message_list.append(rainbow_dc.SubMessage(foo='Foo in a list', bar='Candybar'))
        dc8.message_list.append(rainbow_dc.SubMessage(foo='a Fool in a list', bar='Tequila bar'))

        dc8.simple_map['dora'] = 'Imamap!'
        dc8.simple_map['diego'] = 'Camera!'
        self.assertEqual(dc8, protoplasm.casting.mkdataclass(rainbow_dc.RainbowMessage,
                                                             simple_field='Green',
                                                             message_field__foo='Blue',
                                                             message_field__bar='Yellow',
                                                             simple_list=['Four', 'Seven', '99'],
                                                             message_list=[castutils.kwdict(foo='Foo in a list',
                                                                                            bar='Candybar'),
                                                                           castutils.kwdict(foo='a Fool in a list',
                                                                                            bar='Tequila bar')],
                                                             simple_map__dora='Imamap!',
                                                             simple_map__diego='Camera!'
                                                            ))

        dc9 = rainbow_dc.RainbowMessage()
        dc9.simple_field = 'Green'
        dc9.message_field = rainbow_dc.SubMessage(foo='Blue', bar='Yellow')
        dc9.simple_list.append('Four')
        dc9.simple_list.append('Seven')
        dc9.simple_list.append('99')
        dc9.message_list.append(rainbow_dc.SubMessage(foo='Foo in a list', bar='Candybar'))
        dc9.message_list.append(rainbow_dc.SubMessage(foo='a Fool in a list', bar='Tequila bar'))
        dc9.simple_map['dora'] = 'Imamap!'
        dc9.simple_map['diego'] = 'Camera!'

        dc9.message_map['donald'] = rainbow_dc.SubMessage(foo='duck', bar='trump')

        dc9.message_map['mickey'] = rainbow_dc.SubMessage(foo='mouse')

        expected_dc9 = protoplasm.casting.mkdataclass(rainbow_dc.RainbowMessage,
                                                      simple_field='Green',
                                                      message_field__foo='Blue',
                                                      message_field__bar='Yellow',
                                                      simple_list=['Four', 'Seven', '99'],
                                                      message_list=[castutils.kwdict(foo='Foo in a list',
                                                                                     bar='Candybar'),
                                                                    castutils.kwdict(foo='a Fool in a list',
                                                                                     bar='Tequila bar')],
                                                      simple_map__dora='Imamap!',
                                                      simple_map__diego='Camera!',
                                                      message_map__donald__foo='duck',
                                                      message_map__donald__bar='trump',
                                                      message_map__mickey__foo='mouse'
                                                      )

        self.assertEqual(dc9, expected_dc9)

        dc10 = rainbow_dc.RainbowMessage()
        dc10.simple_field = 'Green'
        dc10.message_field = rainbow_dc.SubMessage(foo='Blue', bar='Yellow')
        dc10.simple_list.append('Four')
        dc10.simple_list.append('Seven')
        dc10.simple_list.append('99')
        dc10.message_list.append(rainbow_dc.SubMessage(foo='Foo in a list', bar='Candybar'))
        dc10.message_list.append(rainbow_dc.SubMessage(foo='a Fool in a list', bar='Tequila bar'))
        dc10.simple_map['dora'] = 'Imamap!'
        dc10.simple_map['diego'] = 'Camera!'

        dc10.message_map['mickey'] = rainbow_dc.SubMessage(foo='mouse')
        dc10.message_map['donald'] = rainbow_dc.SubMessage(foo='duck', bar='trump')

        self.assertEqual(dc10, protoplasm.casting.mkdataclass(rainbow_dc.RainbowMessage,
                                                              simple_map__diego='Camera!',
                                                              simple_field='Green',
                                                              message_field__foo='Blue',
                                                              message_field__bar='Yellow',
                                                              simple_list=['Four', 'Seven', '99'],
                                                              message_list=[castutils.kwdict(foo='Foo in a list',
                                                                                             bar='Candybar'),
                                                                            castutils.kwdict(foo='a Fool in a list',
                                                                                             bar='Tequila bar')],
                                                              simple_map__dora='Imamap!',
                                                              message_map__donald__foo='duck',
                                                              message_map__donald__bar='trump',
                                                              message_map__mickey__foo='mouse'
                                                             ))

    def test_import_dataclass(self):
        from sandbox.test import rainbow_pb2
        from sandbox.test import rainbow_dc

        self.assertEqual(rainbow_dc.SubMessage, castutils.import_dataclass_by_proto(rainbow_pb2.SubMessage))
        self.assertEqual(rainbow_dc.SubMessage, castutils.import_dataclass_by_proto(rainbow_pb2.SubMessage()))
        self.assertEqual(rainbow_dc.RainbowMessage, castutils.import_dataclass_by_proto(rainbow_pb2.RainbowMessage))
        self.assertEqual(rainbow_dc.RainbowMessage, castutils.import_dataclass_by_proto(rainbow_pb2.RainbowMessage()))
        self.assertEqual(castutils.import_dataclass_by_proto(rainbow_pb2.RainbowMessage()), castutils.import_dataclass_by_proto(rainbow_pb2.RainbowMessage()))
        self.assertEqual(castutils.import_dataclass_by_proto(rainbow_pb2.RainbowMessage), castutils.import_dataclass_by_proto(rainbow_pb2.RainbowMessage()))
        self.assertNotEqual(castutils.import_dataclass_by_proto(rainbow_pb2.SubMessage), castutils.import_dataclass_by_proto(rainbow_pb2.RainbowMessage))

    def test_timestamp_mkproto(self):
        from sandbox.test import rainbow_pb2

        ts1 = '2012-07-03T14:50:51.654321Z'
        dt1 = datetime.datetime(2012, 7, 3, 14, 50, 51, 654321)
        self.assertEqual(ts1, dt1.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts2 = '2010-04-09T22:38:39.009870Z'
        dt2 = datetime.datetime(2010, 4, 9, 22, 38, 39, 9870)
        self.assertEqual(ts2, dt2.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts3 = '1979-07-06T14:30:00.000000Z'
        dt3 = datetime.datetime(1979, 7, 6, 14, 30)
        self.assertEqual(ts3, dt3.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts4 = '2049-01-01T12:00:00.000000Z'
        dt4 = datetime.datetime(2049, 1, 1, 12)
        self.assertEqual(ts4, dt4.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts5 = '2049-01-01T00:00:00.000000Z'
        dt5 = datetime.datetime(2049, 1, 1)
        self.assertEqual(ts5, dt5.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        p_expect = rainbow_pb2.TimestampMessage()
        p_expect.my_timestamp.FromJsonString(ts1)
        p_expect.my_timestamp_list.add().FromJsonString(ts2)
        p_expect.my_timestamp_list.add().FromJsonString(ts3)
        p_expect.my_timestamp_map['noon'].FromJsonString(ts4)
        p_expect.my_timestamp_map['midnight'].FromJsonString(ts5)

        self.assertEqual(p_expect, protoplasm.casting.mkproto(rainbow_pb2.TimestampMessage, my_timestamp=ts1,
                                                                                            my_timestamp_list=[ts2, ts3],
                                                                                            my_timestamp_map__noon=ts4,
                                                                                            my_timestamp_map__midnight=ts5))

    def test_timestamp_mkdataclass(self):
        from sandbox.test import rainbow_dc

        ts1 = '2012-07-03T14:50:51.654321Z'
        dt1 = datetime.datetime(2012, 7, 3, 14, 50, 51, 654321)
        self.assertEqual(ts1, dt1.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts2 = '2010-04-09T22:38:39.009870Z'
        dt2 = datetime.datetime(2010, 4, 9, 22, 38, 39, 9870)
        self.assertEqual(ts2, dt2.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts3 = '1979-07-06T14:30:00.000000Z'
        dt3 = datetime.datetime(1979, 7, 6, 14, 30)
        self.assertEqual(ts3, dt3.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts4 = '2049-01-01T12:00:00.000000Z'
        dt4 = datetime.datetime(2049, 1, 1, 12)
        self.assertEqual(ts4, dt4.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts5 = '2049-01-01T00:00:00.000000Z'
        dt5 = datetime.datetime(2049, 1, 1)
        self.assertEqual(ts5, dt5.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        dc_expect = rainbow_dc.TimestampMessage()
        dc_expect.my_timestamp = dt1
        dc_expect.my_timestamp_list.append(dt2)
        dc_expect.my_timestamp_list.append(dt3)
        dc_expect.my_timestamp_map['noon'] = dt4
        dc_expect.my_timestamp_map['midnight'] = dt5

        self.assertEqual(dc_expect, protoplasm.casting.mkdataclass(rainbow_dc.TimestampMessage,
                                                                   my_timestamp=dt1,
                                                                   my_timestamp_list=[dt2, dt3],
                                                                   my_timestamp_map__noon=dt4,
                                                                   my_timestamp_map__midnight=dt5))

    def test_byte_mkproto(self):
        from sandbox.test import rainbow_pb2

        as_str_1 = 'Þórður Matthíasson'
        as_bytes_1 = as_str_1.encode('utf-8')
        as_base64_1 = base64.encodebytes(as_bytes_1).decode('utf-8').strip()

        as_str_2 = '♡'
        as_bytes_2 = as_str_2.encode('utf-8')
        as_base64_2 = base64.encodebytes(as_bytes_2).decode('utf-8').strip()

        as_str_3 = '♠'
        as_bytes_3 = as_str_3.encode('utf-8')
        as_base64_3 = base64.encodebytes(as_bytes_3).decode('utf-8').strip()

        as_str_4 = '♢'
        as_bytes_4 = as_str_4.encode('utf-8')
        as_base64_4 = base64.encodebytes(as_bytes_4).decode('utf-8').strip()

        as_str_5 = '♣'
        as_bytes_5 = as_str_5.encode('utf-8')
        as_base64_5 = base64.encodebytes(as_bytes_5).decode('utf-8').strip()

        p_expect = rainbow_pb2.BytesMessage()
        p_expect.my_bytes = as_bytes_1
        p_expect.my_bytes_list.append(as_bytes_2)
        p_expect.my_bytes_list.append(as_bytes_3)
        p_expect.my_bytes_map['zero'] = as_bytes_4
        p_expect.my_bytes_map['one'] = as_bytes_5

        self.assertEqual(p_expect, protoplasm.casting.mkproto(rainbow_pb2.BytesMessage,
                                                              my_bytes=as_base64_1,
                                                              my_bytes_list=[as_base64_2, as_base64_3],
                                                              my_bytes_map__zero=as_base64_4,
                                                              my_bytes_map__one=as_base64_5))

    def test_byte_mkdataclass(self):
        from sandbox.test import rainbow_dc

        as_str_1 = 'Þórður Matthíasson'
        as_bytes_1 = as_str_1.encode('utf-8')
        as_base64_1 = base64.encodebytes(as_bytes_1).decode('utf-8').strip()

        as_str_2 = '♡'
        as_bytes_2 = as_str_2.encode('utf-8')
        as_base64_2 = base64.encodebytes(as_bytes_2).decode('utf-8').strip()

        as_str_3 = '♠'
        as_bytes_3 = as_str_3.encode('utf-8')
        as_base64_3 = base64.encodebytes(as_bytes_3).decode('utf-8').strip()

        as_str_4 = '♢'
        as_bytes_4 = as_str_4.encode('utf-8')
        as_base64_4 = base64.encodebytes(as_bytes_4).decode('utf-8').strip()

        as_str_5 = '♣'
        as_bytes_5 = as_str_5.encode('utf-8')
        as_base64_5 = base64.encodebytes(as_bytes_5).decode('utf-8').strip()

        dc_expect = rainbow_dc.BytesMessage()
        dc_expect.my_bytes = as_bytes_1
        dc_expect.my_bytes_list.append(as_bytes_2)
        dc_expect.my_bytes_list.append(as_bytes_3)
        dc_expect.my_bytes_map['zero'] = as_bytes_4
        dc_expect.my_bytes_map['one'] = as_bytes_5

        self.assertEqual(dc_expect, protoplasm.casting.mkdataclass(rainbow_dc.BytesMessage,
                                                                   my_bytes=as_bytes_1,
                                                                   my_bytes_list=[as_bytes_2, as_base64_3],
                                                                   my_bytes_map__zero=as_bytes_4,
                                                                   my_bytes_map__one=as_base64_5))

    def test_enum_mkproto(self):
        from sandbox.test import enums_pb2

        p_expect = enums_pb2.WithExternalEnum()
        p_expect.my_enum = enums_pb2.TWO

        p_expect.my_enum_list.append(enums_pb2.ONE)
        p_expect.my_enum_list.append(enums_pb2.THREE)

        p_expect.my_enum_map['one'] = enums_pb2.ONE
        p_expect.my_enum_map['two'] = enums_pb2.TWO

        p_expect.my_alias_enum = enums_pb2.SIX

        p_expect.my_alias_enum_list.append(enums_pb2.FOUR)
        p_expect.my_alias_enum_list.append(enums_pb2.FIVE)
        p_expect.my_alias_enum_list.append(enums_pb2.FJORIR)

        p_expect.my_alias_enum_map['six'] = enums_pb2.SIX
        p_expect.my_alias_enum_map['sex'] = enums_pb2.SEX
        p_expect.my_alias_enum_map['fimm'] = enums_pb2.FIVE

        self.assertEqual(p_expect, protoplasm.casting.mkproto(enums_pb2.WithExternalEnum,
                                                              my_enum=2,
                                                              my_alias_enum=3,
                                                              my_enum_list=[1, 3],
                                                              my_alias_enum_list=[1, 2, 1],
                                                              my_enum_map__one=1,
                                                              my_enum_map__two=2,
                                                              my_alias_enum_map__six=3,
                                                              my_alias_enum_map__sex=3,
                                                              my_alias_enum_map__fimm=2))
        self.assertEqual(p_expect, protoplasm.casting.mkproto(enums_pb2.WithExternalEnum,
                                                              my_enum=enums_pb2.TWO,
                                                              my_alias_enum=enums_pb2.SIX,
                                                              my_enum_list=[enums_pb2.ONE, enums_pb2.THREE],
                                                              my_alias_enum_list=[enums_pb2.FOUR, enums_pb2.FIVE, enums_pb2.FJORIR],
                                                              my_enum_map__one=enums_pb2.ONE,
                                                              my_enum_map__two=enums_pb2.TWO,
                                                              my_alias_enum_map__six=enums_pb2.SIX,
                                                              my_alias_enum_map__sex=enums_pb2.SEX,
                                                              my_alias_enum_map__fimm=enums_pb2.FIMM))

        p_expect2 = enums_pb2.WithInternalEnum()
        p_expect2.my_internal_enum = enums_pb2.WithInternalEnum.SIX

        p_expect2.my_internal_enum_list.append(enums_pb2.WithInternalEnum.FIVE)
        p_expect2.my_internal_enum_list.append(enums_pb2.WithInternalEnum.FOUR)

        p_expect2.my_internal_enum_map['no5'] = enums_pb2.WithInternalEnum.FIVE
        p_expect2.my_internal_enum_map['no6'] = enums_pb2.WithInternalEnum.SIX

        p_expect2.my_internal_alias_enum = enums_pb2.WithInternalEnum.SEVEN

        p_expect2.my_internal_alias_enum_list.append(enums_pb2.WithInternalEnum.SJO)
        p_expect2.my_internal_alias_enum_list.append(enums_pb2.WithInternalEnum.ATTA)
        p_expect2.my_internal_alias_enum_list.append(enums_pb2.WithInternalEnum.EIGHT)

        p_expect2.my_internal_alias_enum_map['no9'] = enums_pb2.WithInternalEnum.NIU
        p_expect2.my_internal_alias_enum_map['no9B'] = enums_pb2.WithInternalEnum.NINE
        p_expect2.my_internal_alias_enum_map['default'] = enums_pb2.WithInternalEnum.ZERO

        self.assertEqual(p_expect2, protoplasm.casting.mkproto(enums_pb2.WithInternalEnum,
                                                               my_internal_enum=enums_pb2.WithInternalEnum.SIX,
                                                               my_internal_alias_enum=enums_pb2.WithInternalEnum.SEVEN,
                                                               my_internal_enum_list=[enums_pb2.WithInternalEnum.FIVE,
                                                                                      enums_pb2.WithInternalEnum.FOUR],
                                                               my_internal_alias_enum_list=[enums_pb2.WithInternalEnum.SJO,
                                                                                            enums_pb2.WithInternalEnum.ATTA,
                                                                                            enums_pb2.WithInternalEnum.EIGHT],
                                                               my_internal_enum_map__no5=enums_pb2.WithInternalEnum.FIVE,
                                                               my_internal_enum_map__no6=enums_pb2.WithInternalEnum.SIX,
                                                               my_internal_alias_enum_map__no9=enums_pb2.WithInternalEnum.NIU,
                                                               my_internal_alias_enum_map__no9B=enums_pb2.WithInternalEnum.NINE,
                                                               my_internal_alias_enum_map__default=enums_pb2.WithInternalEnum.ZERO))

        self.assertEqual(p_expect2, protoplasm.casting.mkproto(enums_pb2.WithInternalEnum,
                                                               my_internal_enum=6,
                                                               my_internal_alias_enum=7,
                                                               my_internal_enum_list=[5, 4],
                                                               my_internal_alias_enum_list=[7, 8, 8],
                                                               my_internal_enum_map__no5=5,
                                                               my_internal_enum_map__no6=6,
                                                               my_internal_alias_enum_map__no9=9,
                                                               my_internal_alias_enum_map__no9B=9,
                                                               my_internal_alias_enum_map__default=0))

    def test_enum_mkdataclass(self):
        from sandbox.test import enums_dc

        dc_expect = enums_dc.WithExternalEnum()
        dc_expect.my_enum = enums_dc.TWO

        dc_expect.my_enum_list.append(enums_dc.ONE)
        dc_expect.my_enum_list.append(enums_dc.THREE)

        dc_expect.my_enum_map['one'] = enums_dc.ONE
        dc_expect.my_enum_map['two'] = enums_dc.TWO

        dc_expect.my_alias_enum = enums_dc.SIX

        dc_expect.my_alias_enum_list.append(enums_dc.FOUR)
        dc_expect.my_alias_enum_list.append(enums_dc.FIVE)
        dc_expect.my_alias_enum_list.append(enums_dc.FJORIR)

        dc_expect.my_alias_enum_map['six'] = enums_dc.SIX
        dc_expect.my_alias_enum_map['sex'] = enums_dc.SEX
        dc_expect.my_alias_enum_map['fimm'] = enums_dc.FIVE

        self.assertEqual(dc_expect, protoplasm.casting.mkdataclass(enums_dc.WithExternalEnum,
                                                                   my_enum=2,
                                                                   my_alias_enum=3,
                                                                   my_enum_list=[1, 3],
                                                                   my_alias_enum_list=[1, 2, 1],
                                                                   my_enum_map__one=1,
                                                                   my_enum_map__two=2,
                                                                   my_alias_enum_map__six=3,
                                                                   my_alias_enum_map__sex=3,
                                                                   my_alias_enum_map__fimm=2))
        self.assertEqual(dc_expect, protoplasm.casting.mkdataclass(enums_dc.WithExternalEnum,
                                                                   my_enum=enums_dc.TWO,
                                                                   my_alias_enum=enums_dc.SIX,
                                                                   my_enum_list=[enums_dc.ONE, enums_dc.THREE],
                                                                   my_alias_enum_list=[enums_dc.FOUR, enums_dc.FIVE,
                                                                                       enums_dc.FJORIR],
                                                                   my_enum_map__one=enums_dc.ONE,
                                                                   my_enum_map__two=enums_dc.TWO,
                                                                   my_alias_enum_map__six=enums_dc.SIX,
                                                                   my_alias_enum_map__sex=enums_dc.SEX,
                                                                   my_alias_enum_map__fimm=enums_dc.FIMM))

        dc_expect2 = enums_dc.WithInternalEnum()
        dc_expect2.my_internal_enum = enums_dc.WithInternalEnum.SIX

        dc_expect2.my_internal_enum_list.append(enums_dc.WithInternalEnum.FIVE)
        dc_expect2.my_internal_enum_list.append(enums_dc.WithInternalEnum.FOUR)

        dc_expect2.my_internal_enum_map['no5'] = enums_dc.WithInternalEnum.FIVE
        dc_expect2.my_internal_enum_map['no6'] = enums_dc.WithInternalEnum.SIX

        dc_expect2.my_internal_alias_enum = enums_dc.WithInternalEnum.SEVEN

        dc_expect2.my_internal_alias_enum_list.append(enums_dc.WithInternalEnum.SJO)
        dc_expect2.my_internal_alias_enum_list.append(enums_dc.WithInternalEnum.ATTA)
        dc_expect2.my_internal_alias_enum_list.append(enums_dc.WithInternalEnum.EIGHT)

        dc_expect2.my_internal_alias_enum_map['no9'] = enums_dc.WithInternalEnum.NIU
        dc_expect2.my_internal_alias_enum_map['no9B'] = enums_dc.WithInternalEnum.NINE
        dc_expect2.my_internal_alias_enum_map['default'] = enums_dc.WithInternalEnum.ZERO

        self.assertEqual(dc_expect2, protoplasm.casting.mkdataclass(enums_dc.WithInternalEnum,
                                                                    my_internal_enum=enums_dc.WithInternalEnum.SIX,
                                                                    my_internal_alias_enum=enums_dc.WithInternalEnum.SEVEN,
                                                                    my_internal_enum_list=[enums_dc.WithInternalEnum.FIVE,
                                                                                           enums_dc.WithInternalEnum.FOUR],
                                                                    my_internal_alias_enum_list=[
                                                                        enums_dc.WithInternalEnum.SJO,
                                                                        enums_dc.WithInternalEnum.ATTA,
                                                                        enums_dc.WithInternalEnum.EIGHT],
                                                                    my_internal_enum_map__no5=enums_dc.WithInternalEnum.FIVE,
                                                                    my_internal_enum_map__no6=enums_dc.WithInternalEnum.SIX,
                                                                    my_internal_alias_enum_map__no9=enums_dc.WithInternalEnum.NIU,
                                                                    my_internal_alias_enum_map__no9B=enums_dc.WithInternalEnum.NINE,
                                                                    my_internal_alias_enum_map__default=enums_dc.WithInternalEnum.ZERO))

        self.assertEqual(dc_expect2, protoplasm.casting.mkdataclass(enums_dc.WithInternalEnum,
                                                                    my_internal_enum=6,
                                                                    my_internal_alias_enum=7,
                                                                    my_internal_enum_list=[5, 4],
                                                                    my_internal_alias_enum_list=[7, 8, 8],
                                                                    my_internal_enum_map__no5=5,
                                                                    my_internal_enum_map__no6=6,
                                                                    my_internal_alias_enum_map__no9=9,
                                                                    my_internal_alias_enum_map__no9B=9,
                                                                    my_internal_alias_enum_map__default=0))



__all__ = [
    'BaseGrpcServicer',
]
import grpc  # noqa

from protoplasm.structs import *
from .dataclass_bases import *
from ._methwrap import *


class BaseGrpcServicer:
    def __init__(self, implementation, *args, **kwargs):
        self.impl = implementation

    def add_to_server(self, server):
        raise NotImplementedError()

    def _forward_to_impl(self, data_cls: Type[T_DCB], impl_method: Callable,
                         request: GeneratedProtocolMessageType,
                         context: grpc.ServicerContext) -> GeneratedProtocolMessageType:
        """Alias to _forward_to_unary_unary_impl for backwards compatibility"""
        return self._forward_to_unary_unary_impl(data_cls, impl_method, request, context)

    def _forward_to_unary_unary_impl(self, data_cls: Type[T_DCB], impl_method: Callable,
                                     request: GeneratedProtocolMessageType,
                                     context: grpc.ServicerContext) -> GeneratedProtocolMessageType:
        return PlasmMethodWrapper(self, data_cls, impl_method, request, context,
                                  MethodIoType.UNARY_IN_UNARY_OUT).call()

    def _forward_to_unary_stream_impl(self, data_cls: Type[T_DCB], impl_method: Callable,
                                      request: GeneratedProtocolMessageType,
                                      context: grpc.ServicerContext) -> Iterable[GeneratedProtocolMessageType]:
        return PlasmMethodWrapper(self, data_cls, impl_method, request, context,
                                  MethodIoType.UNARY_IN_STREAM_OUT).call()

    def _forward_to_stream_unary_impl(self, data_cls: Type[T_DCB], impl_method: Callable,
                                      request_iterator: Iterable[GeneratedProtocolMessageType],
                                      context: grpc.ServicerContext):
        return PlasmMethodWrapper(self, data_cls, impl_method, request_iterator, context,
                                  MethodIoType.STREAM_IN_UNARY_OUT).call()

    def _forward_to_stream_stream_impl(self, data_cls: Type[T_DCB], impl_method: Callable,
                                       request_iterator: Iterable[GeneratedProtocolMessageType],
                                       context: grpc.ServicerContext) -> Iterable[GeneratedProtocolMessageType]:
        return PlasmMethodWrapper(self, data_cls, impl_method, request_iterator, context,
                                  MethodIoType.STREAM_IN_STREAM_OUT).call()



__all__ = [
    'BaseContext',
]


class BaseContext(object):
    def __init__(self, wrapped_context, *args, **kwargs):
        self.wrapped_context = wrapped_context
        self.metadata = {}

    def get_origin_ip(self, default=None):
        raise NotImplementedError('BaseContext needs to be extended to account for different invocation contexts')

    def get_meta(self, name, default=None):
        raise NotImplementedError('BaseContext needs to be extended to account for different invocation contexts')



__all__ = [
    'Unknown',
    'NotFound',
    'InvalidArgument',
    'Unimplemented',
    'PermissionDenied',
    'Unauthenticated',
    'Unavailable',
    'DeadlineExceeded',
    'AlreadyExists',
    'FailedPrecondition',
    'DataLoss',
    'ResourceExhausted',
    'OutOfRange',
    'Internal',
    'Cancelled',
    'Aborted',
    'ClientApiError',
    'PreRequestError',
    'PostRequestError',
]

import grpc
from ._base import *


class Unknown(ServiceApiException):
    def __init__(self, details='unknown', ref_code=None, should_log=True):
        """For example, this error may be returned when a Status value received
        from another address space belongs to an error-space that is not known
        in this address space. Also errors raised by APIs that do not return
        enough error information may be converted to this error.
        """
        self.status_code = grpc.StatusCode.UNKNOWN
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class NotFound(ServiceApiException):
    def __init__(self, details='not found', ref_code=None, should_log=False):
        """Some requested entity was not found.
        """
        self.status_code = grpc.StatusCode.NOT_FOUND
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class InvalidArgument(ServiceApiException):
    def __init__(self, details='invalid argument', ref_code=None, should_log=False):
        """The client specified an invalid argument.
        """
        self.status_code = grpc.StatusCode.INVALID_ARGUMENT
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class Unimplemented(ServiceApiException):
    def __init__(self, details='unimplemented', ref_code=None, should_log=False):
        """The operation is not implemented or is not supported/enabled in this
        service.
        """
        self.status_code = grpc.StatusCode.UNIMPLEMENTED
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class PermissionDenied(ServiceApiException):
    def __init__(self, details='permission denied', ref_code=None, should_log=True):
        """The caller does not have permission to execute the specified
        operation. PERMISSION_DENIED must not be used for rejections caused by
        exhausting some resource (use RESOURCE_EXHAUSTED instead for those
        errors). PERMISSION_DENIED must not be used if the caller can not be
        identified (use UNAUTHENTICATED instead for those errors). This error
        code does not imply the request is valid or the requested entity exists
        or satisfies other pre-conditions.
        """
        self.status_code = grpc.StatusCode.PERMISSION_DENIED
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class Unauthenticated(ServiceApiException):
    def __init__(self, details='unauthenticated', ref_code=None, should_log=False):
        """The request is not authenticated but needs to be.
        """
        self.status_code = grpc.StatusCode.UNAUTHENTICATED
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class Unavailable(ServiceApiException):
    def __init__(self, details='unavailable', ref_code=None, should_log=False):
        """The service is currently unavailable. This is most likely a transient
        condition, which can be corrected by retrying with a backoff.
        """
        self.status_code = grpc.StatusCode.UNAVAILABLE
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class DeadlineExceeded(ServiceApiException):
    def __init__(self, details='deadline exceeded', ref_code=None, should_log=False):
        """The deadline expired before the operation could complete. For
        operations that change the state of the system, this error may be
        returned even if the operation has completed successfully. For example,
        a successful response from a server could have been delayed long enough
        for the deadline to expire.
        """
        self.status_code = grpc.StatusCode.DEADLINE_EXCEEDED
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class AlreadyExists(ServiceApiException):
    def __init__(self, details='already exists', ref_code=None, should_log=False):
        """The entity that a client attempted to create already exists.
        """
        self.status_code = grpc.StatusCode.ALREADY_EXISTS
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class FailedPrecondition(ServiceApiException):
    def __init__(self, details='failed precondition', ref_code=None, should_log=False):
        """The operation was rejected because the system is not in a state
        required for the operation's execution. For example, the directory to be
        deleted is non-empty, an rmdir operation is applied to a non-directory,
        etc.
        """
        self.status_code = grpc.StatusCode.FAILED_PRECONDITION
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class DataLoss(ServiceApiException):
    def __init__(self, details='data loss', ref_code=None, should_log=True):
        """Unrecoverable data loss or corruption.
        """
        self.status_code = grpc.StatusCode.DATA_LOSS
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class ResourceExhausted(ServiceApiException):
    def __init__(self, details='resource exhausted', ref_code=None, should_log=True):
        """Some resource has been exhausted, perhaps a per-user quota, or
        perhaps the entire file system is out of space.
        """
        self.status_code = grpc.StatusCode.RESOURCE_EXHAUSTED
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class OutOfRange(ServiceApiException):
    def __init__(self, details='out of range', ref_code=None, should_log=False):
        """The operation was attempted past the valid range.
        """
        self.status_code = grpc.StatusCode.OUT_OF_RANGE
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class Internal(ServiceApiException):
    def __init__(self, details='internal', ref_code=None, should_log=True):
        """Internal errors. This means that some invariants expected by the
        underlying system have been broken. This error code is reserved for
        serious errors.
        """
        self.status_code = grpc.StatusCode.INTERNAL
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class Cancelled(ServiceApiException):
    def __init__(self, details='cancelled', ref_code=None, should_log=False):
        """The operation was cancelled, typically by the caller.
        """
        self.status_code = grpc.StatusCode.CANCELLED
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class Aborted(ServiceApiException):
    def __init__(self, details='aborted', ref_code=None, should_log=False):
        """The operation was aborted, typically due to a concurrency issue such
        as a sequencer check failure or transaction abort.
        """
        self.status_code = grpc.StatusCode.ABORTED
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class ClientApiError(ServiceApiException):
    def __init__(self, details='client_error', ref_code=None, should_log=True):
        """Error raised by a gRPC client when non-gRPX protoplasm errors occur.
        """
        self.status_code = grpc.StatusCode.UNKNOWN
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class PreRequestError(ClientApiError):
    def __init__(self, details='pre_request_error', ref_code=None, should_log=True):
        """Error raised by a gRPC client when exceptions occur before the actual
        gRPC request (e.g. while casting).
        """
        self.status_code = grpc.StatusCode.FAILED_PRECONDITION
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code


class PostRequestError(ClientApiError):
    def __init__(self, details='post_request_error', ref_code=None, should_log=True):
        """Error raised by a gRPC client when exceptions occur AFTER the actual
        gRPC request (e.g. while casting).
        """
        self.status_code = grpc.StatusCode.CANCELLED
        self.details = details
        self.should_log = should_log
        self.ref_code = ref_code



__all__ = [
    'BaseDictator',
    'TimestampDictator',
    'ByteDictator',
    'EnumDictator',
    'LongDictator',
    'DurationDictator',
    'AnyDictator',
    'StructDictator',
]
import datetime
import base64
import dataclasses
from ccptools import dtu
import enum
import collections

from protoplasm.structs import *
from protoplasm.casting import castutils


class BaseDictator:
    @classmethod
    def to_dict_value(cls, dc_value: Any, field: dataclasses.Field, parent: DataclassBase) -> Any:
        """Casts data from whatever a dataclass stores to a value that protobuf
        messages can parse from a dict.

        :param dc_value: Dataclass value
        :type dc_value: Any
        :param field: The dataclass field descriptor the value comes from
        :type field: dataclasses.Field
        :param parent: The dataclass the field belongs to
        :type parent: object
        :return: A value that the protobuf ParseDict function can use
        :rtype: Any
        """
        return dc_value

    @classmethod
    def from_dict_value(cls, proto_value: Any, field: dataclasses.Field, parent_type: Type[DataclassBase]) -> Any:
        """Casts data from a dict version of a protobuf message into whatever
        value the corresponding dataclass uses.

        :param proto_value: Protobuf dict value
        :type proto_value: Any
        :param field: The dataclass field descriptor the value is going to
        :type field: dataclasses.Field
        :param parent_type: The dataclass the field belongs to
        :type parent_type: object
        :return: A value that the dataclass uses
        :rtype: Any
        """
        return proto_value


class TimestampDictator(BaseDictator):
    @classmethod
    def to_dict_value(cls, dc_value: Optional[datetime.datetime],
                      field: dataclasses.Field, parent: DataclassBase) -> Optional[str]:
        """Casts data from a dataclass datetime values to iso string (with
        tailing Z).

        Format: %Y-%m-%dT%H:%M:%S.%fZ

        :param dc_value: Dataclass value of a python datetime
        :type dc_value: datetime.datetime | None
        :param field: The dataclass field descriptor the value comes from
        :type field: dataclasses.Field
        :param parent: The dataclass the field belongs to
        :type parent: object
        :return: Datetime as an iso string (with trailing Z)
        :rtype: str | None
        """
        if not dc_value:
            return None

        if not isinstance(dc_value, datetime.datetime):
            raise TypeError(f'TimestampDictator.to_dict_value expected datetime, got {type(dc_value)}')

        if dc_value.microsecond:
            return dc_value.strftime('%Y-%m-%dT%H:%M:%S.%fZ')

        return dc_value.strftime('%Y-%m-%dT%H:%M:%SZ')

    @classmethod
    def from_dict_value(cls, proto_value: Optional[str],
                        field: dataclasses.Field, parent_type: Type[DataclassBase]) -> Optional[datetime.datetime]:
        """Casts a timestamp value from a iso-string (with trailing Z) to a
        datetime.

        Format: %Y-%m-%dT%H:%M:%S.%fZ

        :param proto_value: Protobuf dict value of an iso-string (with trailing Z)
        :type proto_value: str | None
        :param field: The dataclass field descriptor the value is going to
        :type field: dataclasses.Field
        :param parent_type: The dataclass the field belongs to
        :type parent_type: object
        :return: A python datetime value
        :rtype: datetime.datetime | None
        """
        if not proto_value:
            return None

        val = dtu.any_to_datetime(proto_value, None)
        if val is None:
            raise ValueError(f'TimestampDictator.from_dict_value failed to convert "{proto_value!r}" to datetime')

        return val


class ByteDictator(BaseDictator):
    @classmethod
    def to_dict_value(cls, dc_value: Optional[bytes],
                      field: dataclasses.Field, parent: DataclassBase) -> str:
        """Casts data from python bytes to base64 encoded string (bytes)

        :param dc_value: Dataclass value of python bytes
        :type dc_value: bytes | None
        :param field: The dataclass field descriptor the value comes from
        :type field: dataclasses.Field
        :param parent: The dataclass the field belongs to
        :type parent: object
        :return: A base64 encoded string (bytes) version of the bytes
        :rtype: bytes
        """
        if not dc_value:
            return ''

        if not isinstance(dc_value, bytes):
            raise TypeError(f'ByteDictator.to_dict_value expected bytes, got {type(dc_value)}')

        return base64.encodebytes(dc_value).decode('utf-8').strip()

    @classmethod
    def from_dict_value(cls, proto_value: Optional[Union[str, bytes]],
                        field: dataclasses.Field, parent_type: Type[DataclassBase]) -> bytes:
        """Casts a base64 encodes string of bytes into python bytes from a iso-string (with trailing Z) to a
        datetime.

        :param proto_value: Protobuf dict value of a base64 endoced string
        :type proto_value: str | bytes | None
        :param field: The dataclass field descriptor the value is going to
        :type field: dataclasses.Field
        :param parent_type: The dataclass the field belongs to
        :type parent_type: object
        :return: A python bytes value
        :rtype: bytes | None
        """
        if not proto_value:
            return b''

        if isinstance(proto_value, bytes):
            return proto_value

        if isinstance(proto_value, str):
            return base64.decodebytes(proto_value.encode('utf-8'))


class EnumDictator(BaseDictator):
    @classmethod
    def to_dict_value(cls, dc_value: Optional[Union[int, enum.IntEnum]],
                      field: dataclasses.Field, parent: DataclassBase) -> int:
        """Casts data from python enum type to plain int for Protobuf

        :param dc_value: Dataclass value of python enum
        :type dc_value: int | enum.IntEnum | None
        :param field: The dataclass field descriptor the value comes from
        :type field: dataclasses.Field
        :param parent: The dataclass the field belongs to
        :type parent: object
        :return: Integer value of the enum
        :rtype: int | None
        """
        if dc_value is None:
            return 0

        if isinstance(dc_value, enum.IntEnum):
            return dc_value.value

        return dc_value

    @classmethod
    def from_dict_value(cls, proto_value: Optional[Union[int, str]],
                        field: dataclasses.Field, parent_type: Type[DataclassBase]) -> enum.IntEnum:
        """Casts an integer or string from a Protobuf enum into a Python enum object.

        :param proto_value: Protobuf int or string version of an enum
        :type proto_value: int | str | None
        :param field: The dataclass field descriptor the value is going to
        :type field: dataclasses.Field
        :param parent_type: The dataclass the field belongs to
        :type parent_type: object
        :return: A Python enum.IntEnum object
        :rtype: enum.IntEnum | None
        """
        if proto_value is None:
            proto_value = 0

        enum_type: enum.EnumMeta = castutils.get_dc_field_obj_type(field, parent_type)

        if isinstance(proto_value, str):
            return enum_type[proto_value]

        return enum_type(proto_value)


class LongDictator:
    @classmethod
    def to_dict_value(cls, dc_value: Union[str, int], field: dataclasses.Field, parent: DataclassBase) -> str:
        """

        :param dc_value: Dataclass value
        :type dc_value: Any
        :param field: The dataclass field descriptor the value comes from
        :type field: dataclasses.Field
        :param parent: The dataclass the field belongs to
        :type parent: object
        :return: A value that the protobuf ParseDict function can use
        :rtype: Any
        """
        if dc_value is None:
            return '0'
        return str(dc_value)

    @classmethod
    def from_dict_value(cls, proto_value: Union[str, int],
                        field: dataclasses.Field, parent_type: Type[DataclassBase]) -> int:
        """

        :param proto_value: Protobuf dict value
        :type proto_value: Any
        :param field: The dataclass field descriptor the value is going to
        :type field: dataclasses.Field
        :param parent_type: The dataclass the field belongs to
        :type parent_type: object
        :return: A value that the dataclass uses
        :rtype: Any
        """
        if proto_value is None:
            return 0
        return int(proto_value)


class DurationDictator:
    @classmethod
    def to_dict_value(cls, dc_value: datetime.timedelta,
                      field: dataclasses.Field, parent: DataclassBase) -> Optional[str]:
        """google.protobuf.duration
        datetime.timedelta -> '123.456789s'
        """
        if dc_value is None:
            return None

        return f'{dc_value.total_seconds():f}s'

    @classmethod
    def from_dict_value(cls, proto_value: Union[str, int, float],
                        field: dataclasses.Field, parent_type: Type[DataclassBase]) -> Optional[datetime.timedelta]:
        """google.protobuf.duration
        '123.456789s' -> datetime.timedelta
        """
        if proto_value is None:
            return None

        if isinstance(proto_value, str):
            proto_value = proto_value.strip().lower()
            if proto_value[-1] == 's':
                proto_value = proto_value[:-1]
            proto_value = float(proto_value)

        return datetime.timedelta(seconds=proto_value)


class AnyDictator:
    @classmethod
    def to_dict_value(cls, dc_value: Optional[DataclassBase],
                      field: dataclasses.Field, parent: DataclassBase) -> Optional[collections.OrderedDict]:
        """google.protobuf.Any
        DataclassBase -> OrderedDict([('@type', '?'), ('key_1', 'value_1'), ('key_n', 'value_n')])
        """
        if dc_value is None:
            return None

        from protoplasm.casting.dictator import dataclass_to_dict
        from protoplasm.casting import castutils

        base_dict = dataclass_to_dict(dc_value)

        ordered_dict = collections.OrderedDict()
        ordered_dict['@type'] = castutils.get_type_url(dc_value.__proto_cls__)

        ordered_dict.update(base_dict)
        return ordered_dict

    @classmethod
    def from_dict_value(cls, proto_value: collections.OrderedDict,
                        field: dataclasses.Field, parent_type: Type[DataclassBase]) -> Optional[DataclassBase]:
        """google.protobuf.Any
        OrderedDict([('@type', '?'), ('key_1', 'value_1'), ('key_n', 'value_n')]) -> DataclassBase
        """
        if proto_value is None or '@type' not in proto_value:
            return None

        from protoplasm.casting import castutils
        from protoplasm.casting.objectifier import dict_to_dataclass

        type_url = proto_value.pop('@type')
        proto_class = castutils.get_proto_descriptor_by_type_url(type_url)._concrete_class  # noqa
        dc_cls = castutils.import_dataclass_by_proto(proto_class)

        return dict_to_dataclass(dc_cls, proto_value)


class StructDictator:
    @classmethod
    def to_dict_value(cls, dc_value: Dict[str, Any],
                      field: dataclasses.Field, parent: DataclassBase) -> Dict[str, Any]:
        if dc_value is None:
            return {}

        return dc_value

    @classmethod
    def from_dict_value(cls, proto_value: Dict[str, Any],
                        field: dataclasses.Field, parent_type: Type[DataclassBase]) -> Optional[Dict[str, Any]]:
        if proto_value is None:
            return {}

        return proto_value



__all__ = [
    'PlasmMethodWrapper',
]

import grpc
import json
import uuid

from ccptools.tpu import casting as typeutils_casting

from protoplasm import casting
from protoplasm import errors
from protoplasm.bases.dataclass_bases import *
from protoplasm.structs import *

if TYPE_CHECKING:
    from .grpc_bases import BaseGrpcServicer

import logging
log = logging.getLogger(__name__)


class PlasmMethodWrapper:
    JSON_SERIALIZER = typeutils_casting.JsonSafeSerializer()

    __slots__ = ('servicer', 'data_cls', 'impl_method', 'request', 'request_iterator', 'context', 'request_dc',
                 'unpacked_args', 'raw_response_iterable', 'raw_response', 'response_dc', 'response_proto',
                 '_ref_code', 'method_io')

    def __init__(self, servicer: 'BaseGrpcServicer', data_cls: Type[T_DCB], impl_method: Callable,
                 request_or_iterator: Union[GeneratedProtocolMessageType, Iterable[GeneratedProtocolMessageType]],
                 context: grpc.ServicerContext, method_io: MethodIoType):
        """Container and worker for processing Servicer (i.e. Server Side)
        requests and responses, packing, unpacking, casting and so on as well as
        handling errors and log-dumping and all that bulky noisy jass that just
        clutters up the actual servicer code.
        """
        self.servicer: 'BaseGrpcServicer' = servicer
        self.data_cls: Type[DataclassBase] = data_cls
        self.impl_method: Callable = impl_method

        self.request: Optional[GeneratedProtocolMessageType] = None
        self.request_iterator: Optional[Iterable[GeneratedProtocolMessageType]] = None

        self.context: grpc.ServicerContext = context
        self.method_io: MethodIoType = method_io

        if self.method_io.is_input_stream():
            self.request_iterator = request_or_iterator
        else:
            self.request = request_or_iterator

        self.request_dc: Optional[DataclassBase] = None
        self.unpacked_args: Optional[Dict] = None
        self.raw_response_iterable: Optional[Iterable[Any]] = None
        self.raw_response: Any = None
        self.response_dc: Optional[DataclassBase] = None
        self.response_proto: Optional[GeneratedProtocolMessageType] = None
        self._ref_code: str = ''

    @property
    def request_dc_iterator(self) -> Iterable[T_DCB]:
        for raw_req in self.request_iterator:
            self.request = raw_req
            self._p2d()
            yield self.request_dc

    @property
    def ref_code(self) -> str:
        if not self._ref_code:
            self._ref_code = str(uuid.uuid4())
        return self._ref_code

    def _get_dump(self, **kwargs) -> dict:
        d = {
            'ref_code': self.ref_code,
            'class': str(self.servicer.__class__),
            'data_cls': str(self.data_cls),
            'impl_method': str(self.impl_method),
            'request': str(self.request),
            'method_io': self.method_io,
        }

        if self.context:
            try:
                d['invocation_metadata'] = dict(self.context.invocation_metadata())  # noqa
            except Exception:
                d['invocation_metadata'] = 'FAILED TO FETCH METADATA!'

        if self.request_dc:
            d['request_dc'] = repr(self.request_dc)
        if self.unpacked_args:
            d['unpacked_args'] = self.unpacked_args
        if self.raw_response:
            d['raw_response'] = repr(self.raw_response)
        if self.response_dc:
            d['response_dc'] = repr(self.response_dc)
        if self.response_proto:
            d['response_proto'] = repr(self.response_proto)

        if kwargs:
            d.update(kwargs)

        return d

    def _get_json_dump(self, **kwargs) -> str:
        return json.dumps(self.JSON_SERIALIZER.serialize(self._get_dump(**kwargs)), indent=4)

    def _abort(self, message: str, error_code, ex):
        self.context.set_trailing_metadata([('ref_code', self.ref_code)])
        log.exception(f'{message}::{ex!r}\n---\n{self._get_json_dump(details=repr(ex))}')
        return self.context.abort(error_code, f'{ex!r};\nref_code={self.ref_code}')

    def _abort_internal(self, ex: errors.ServiceApiException):
        details = [ex.details]
        if ex.should_log:
            details.append(f'ref_code={self.ref_code}')
            self.context.set_trailing_metadata([('ref_code', self.ref_code)])
            log.exception(f'Logging ServiceApiException::{ex!r}\n---\n'
                          f'{self._get_json_dump(status_code=ex.status_code, details=ex.details)}')
        return self.context.abort(ex.status_code, ';\n'.join(details))

    def _p2d(self):
        try:
            self.request_dc = casting.proto_to_dataclass(self.request)
        except Exception as ex:
            return self._abort('pre_request_error;Error while casting proto to dataclass',
                               grpc.StatusCode.FAILED_PRECONDITION, ex)

    def _unpack(self):
        try:
            if self.method_io.is_input_stream():
                self.unpacked_args = {'request_iterator': self.request_dc_iterator}
            else:
                self.unpacked_args = self.request_dc.unpack()
        except Exception as ex:
            return self._abort('pre_request_error;Error while unpacking arguments',
                               grpc.StatusCode.FAILED_PRECONDITION, ex)

    def _add_ctx(self):
        try:
            ctx_name = getattr(self.impl_method, '__takes_context_as__', False)
            if ctx_name:
                self.unpacked_args[ctx_name] = GrpcContext(self.context, self.method_io)
        except Exception as ex:
            return self._abort('pre_request_error;Error while adding context to parameters',
                               grpc.StatusCode.FAILED_PRECONDITION, ex)

    def prepare(self):
        if not self.method_io.is_input_stream():
            self._p2d()
        self._unpack()
        self._add_ctx()

    def _pack(self):
        self.response_dc = None
        try:
            self.response_dc = casting.pack_dataclass_response(self.data_cls, self.raw_response)
        except Exception as ex:
            return self._abort('post_request_error;Error while packing raw response to dataclass',
                               grpc.StatusCode.CANCELLED, ex)

    def _d2p(self):
        self.response_proto = None
        try:
            self.response_proto = casting.dataclass_to_proto(self.response_dc)
        except Exception as ex:
            return self._abort('post_request_error;Error while casting dataclass to proto',
                               grpc.StatusCode.CANCELLED, ex)

    def call(self) -> Union[GeneratedProtocolMessageType,
                            Iterable[GeneratedProtocolMessageType]]:
        # Prepare the call
        self.prepare()

        try:
            if self.method_io.is_output_stream():
                self.raw_response_iterable = self.impl_method(**self.unpacked_args)

            else:
                self.raw_response = self.impl_method(**self.unpacked_args)

        except errors.ServiceApiException as ex:
            return self._abort_internal(ex)

        except Exception as ex:
            return self._abort('Unexpected exception in implementation method',
                               grpc.StatusCode.UNKNOWN, ex)

        if not self.method_io.is_output_stream():
            # Pack up results for unary results
            self._pack()
            self._d2p()
            return self.response_proto

        else:
            return self.iterable_response_proto

    @property
    def wrapped_iterable_raw_response(self) -> Iterable[Any]:
        try:
            for res in self.raw_response_iterable:
                yield res
        except errors.ServiceApiException as ex:
            return self._abort_internal(ex)
        except Exception as ex:
            return self._abort('Unexpected exception in implementation method',
                               grpc.StatusCode.UNKNOWN, ex)

    @property
    def iterable_response_proto(self) -> Iterable[GeneratedProtocolMessageType]:
        for res in self.wrapped_iterable_raw_response:
            self.raw_response = res
            self._pack()
            self._d2p()
            yield self.response_proto

            self.raw_response = None
            self.response_dc = None
            self.response_proto = None



import datetime
import unittest
import base64
import os
import sys

from protoplasm import casting
from protoplasm.casting import dictator
from protoplasm.casting import castutils

import logging
log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

import shutil
import time
HERE = os.path.dirname(__file__)
PROTO_ROOT = os.path.join(HERE, 'res', 'proto')
BUILD_ROOT = os.path.join(HERE, 'res', 'build')


class ProtoToDictTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Remove old stuff...

        build_package = os.path.join(BUILD_ROOT, 'sandbox')
        if os.path.exists(build_package):
            shutil.rmtree(build_package)
            time.sleep(0.1)

        from neobuilder.neobuilder import NeoBuilder

        # Build stuff...
        builder = NeoBuilder(package='sandbox',
                             protopath=PROTO_ROOT,
                             build_root=BUILD_ROOT)
        builder.build()

        # Add build root to path to access its modules
        sys.path.append(BUILD_ROOT)

    def test_proto_to_dict(self):
        from sandbox.test import rainbow_pb2

        expected = {'simple_field': 'Green',
                    'message_field': {
                        'foo': 'Blue',
                        'bar': 'Yellow'
                    },
                    'simple_list': ['Four', 'Seven', '99'],
                    'message_list': [{'foo': 'Foo in a list',
                                      'bar': 'Candybar'},
                                     {'foo': 'a Fool in a list',
                                      'bar': 'Tequila bar'}],
                    'simple_map': {'dora': 'Imamap!',
                                   'diego': 'Camera!'},
                    'message_map': {'mickey': {'foo': 'mouse'},
                                    'donald': {'foo': 'duck',
                                               'bar': 'trump'}}}

        p11 = casting.mkproto(rainbow_pb2.RainbowMessage,
                                         simple_map__diego='Camera!',
                                         simple_field='Green',
                                         message_field__foo='Blue',
                                         message_field__bar='Yellow',
                                         simple_list=['Four', 'Seven', '99'],
                                         message_list=[castutils.kwdict(foo='Foo in a list',
                                                               bar='Candybar'),
                                              castutils.kwdict(foo='a Fool in a list',
                                                               bar='Tequila bar')],
                                         simple_map__dora='Imamap!',
                                         message_map__donald__foo='duck',
                                         message_map__donald__bar='trump',
                                         message_map__mickey__foo='mouse')

        self.assertEqual(expected, dictator.proto_to_dict(p11))

    def test_proto_timestamp_to_dict(self):
        from sandbox.test import rainbow_pb2

        ts1 = '2012-07-03T14:50:51.654321Z'
        dt1 = datetime.datetime(2012, 7, 3, 14, 50, 51, 654321)
        self.assertEqual(ts1, dt1.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts2 = '2010-04-09T22:38:39.009870Z'
        dt2 = datetime.datetime(2010, 4, 9, 22, 38, 39, 9870)
        self.assertEqual(ts2, dt2.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts3 = '1979-07-06T14:30:00Z'
        dt3 = datetime.datetime(1979, 7, 6, 14, 30)
        self.assertEqual(ts3, dt3.strftime('%Y-%m-%dT%H:%M:%SZ'))

        ts4 = '2049-01-01T12:00:00Z'
        dt4 = datetime.datetime(2049, 1, 1, 12)
        self.assertEqual(ts4, dt4.strftime('%Y-%m-%dT%H:%M:%SZ'))

        ts5 = '2049-01-01T00:00:00Z'
        dt5 = datetime.datetime(2049, 1, 1)
        self.assertEqual(ts5, dt5.strftime('%Y-%m-%dT%H:%M:%SZ'))

        expected = {'my_timestamp': ts1,
                    'my_timestamp_list': [ts2, ts3],
                    'my_timestamp_map': {
                        'noon': ts4,
                        'midnight': ts5}}

        p12 = casting.mkproto(rainbow_pb2.TimestampMessage,
                                         my_timestamp=ts1,
                                         my_timestamp_list=[ts2, ts3],
                                         my_timestamp_map__noon=ts4,
                                         my_timestamp_map__midnight=ts5)

        self.assertEqual(expected, dictator.proto_to_dict(p12))

    def test_proto_bytes_to_dict(self):
        from sandbox.test import rainbow_pb2

        as_str_1 = 'Þórður Matthíasson'
        as_bytes_1 = as_str_1.encode('utf-8')
        as_base64_1 = base64.encodebytes(as_bytes_1).decode('utf-8').strip()

        as_str_2 = '♡'
        as_bytes_2 = as_str_2.encode('utf-8')
        as_base64_2 = base64.encodebytes(as_bytes_2).decode('utf-8').strip()

        as_str_3 = '♠'
        as_bytes_3 = as_str_3.encode('utf-8')
        as_base64_3 = base64.encodebytes(as_bytes_3).decode('utf-8').strip()

        as_str_4 = '♢'
        as_bytes_4 = as_str_4.encode('utf-8')
        as_base64_4 = base64.encodebytes(as_bytes_4).decode('utf-8').strip()

        as_str_5 = '♣'
        as_bytes_5 = as_str_5.encode('utf-8')
        as_base64_5 = base64.encodebytes(as_bytes_5).decode('utf-8').strip()

        expected = {'my_bytes': as_base64_1,
                    'my_bytes_list': [as_base64_2, as_base64_3],
                    'my_bytes_map': {
                        'zero': as_base64_4,
                        'one': as_base64_5}}

        p13 = casting.mkproto(rainbow_pb2.BytesMessage,
                                         my_bytes=as_base64_1,
                                         my_bytes_list=[as_base64_2, as_base64_3],
                                         my_bytes_map__zero=as_base64_4,
                                         my_bytes_map__one=as_base64_5)

        self.assertEqual(expected, dictator.proto_to_dict(p13))

    def test_proto_enums_to_dict(self):
        from sandbox.test import enums_pb2

        p1 = enums_pb2.WithExternalEnum()
        p1.my_enum = enums_pb2.TWO

        p1.my_enum_list.append(enums_pb2.ONE)
        p1.my_enum_list.append(enums_pb2.THREE)

        p1.my_enum_map['one'] = enums_pb2.ONE
        p1.my_enum_map['two'] = enums_pb2.TWO

        p1.my_alias_enum = enums_pb2.SIX

        p1.my_alias_enum_list.append(enums_pb2.FOUR)
        p1.my_alias_enum_list.append(enums_pb2.FIVE)
        p1.my_alias_enum_list.append(enums_pb2.FJORIR)

        p1.my_alias_enum_map['six'] = enums_pb2.SIX
        p1.my_alias_enum_map['sex'] = enums_pb2.SEX
        p1.my_alias_enum_map['fimm'] = enums_pb2.FIVE

        expected = {'my_enum': 2,
                    'my_alias_enum': 3,
                    'my_enum_list': [1, 3],
                    'my_alias_enum_list': [1, 2, 1],
                    'my_enum_map': {'one': 1,
                                    'two': 2},
                    'my_alias_enum_map': {'six': 3,
                                          'sex': 3,
                                          'fimm': 2}}

        self.assertEqual(expected, dictator.proto_to_dict(p1))

        p2 = enums_pb2.WithInternalEnum()
        p2.my_internal_enum = enums_pb2.WithInternalEnum.SIX

        p2.my_internal_enum_list.append(enums_pb2.WithInternalEnum.FIVE)
        p2.my_internal_enum_list.append(enums_pb2.WithInternalEnum.FOUR)

        p2.my_internal_enum_map['no5'] = enums_pb2.WithInternalEnum.FIVE
        p2.my_internal_enum_map['no6'] = enums_pb2.WithInternalEnum.SIX

        p2.my_internal_alias_enum = enums_pb2.WithInternalEnum.SEVEN

        p2.my_internal_alias_enum_list.append(enums_pb2.WithInternalEnum.SJO)
        p2.my_internal_alias_enum_list.append(enums_pb2.WithInternalEnum.ATTA)
        p2.my_internal_alias_enum_list.append(enums_pb2.WithInternalEnum.EIGHT)

        p2.my_internal_alias_enum_map['no9'] = enums_pb2.WithInternalEnum.NIU
        p2.my_internal_alias_enum_map['no9B'] = enums_pb2.WithInternalEnum.NINE
        p2.my_internal_alias_enum_map['default'] = enums_pb2.WithInternalEnum.ZERO

        expected2 = {'my_internal_enum': 6,
                     'my_internal_alias_enum': 7,
                     'my_internal_enum_list': [5, 4],
                     'my_internal_alias_enum_list': [7, 8, 8],
                     'my_internal_enum_map': {'no5': 5,
                                              'no6': 6},
                     'my_internal_alias_enum_map': {'no9': 9,
                                                    'no9B': 9,
                                                    'default': 0}}

        self.assertEqual(expected2, dictator.proto_to_dict(p2))

    def test_proto_struct_to_dict(self):
        from sandbox.test import googlestruct_pb2

        expected_a = {
            'my_bool': True,
            'my_float': 4.2,
            'my_null': None,
            'my_dict': {'foo': 'bar', 'you': 'tube'},
            'my_list': [1.0, 3.0, 5.0, 8.0],
            'my_string': 'I am String, hear me spell!',
            'my_int': 42.0
        }

        expected_b = {
            'my_struct': expected_a
        }

        proto_struct = googlestruct_pb2.StructMessage()
        proto_struct.my_struct['my_string'] = 'I am String, hear me spell!'
        proto_struct.my_struct['my_int'] = 42
        proto_struct.my_struct['my_float'] = 4.2
        proto_struct.my_struct['my_null'] = None
        proto_struct.my_struct['my_bool'] = True
        proto_struct.my_struct['my_list'] = [1, 3, 5, 8]
        proto_struct.my_struct['my_dict'] = {'foo': 'bar', 'you': 'tube'}

        self.assertEqual(expected_a, dictator.proto_to_dict(proto_struct.my_struct))

        self.assertEqual(expected_b, dictator.proto_to_dict(proto_struct))


class DataclassToDictTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Remove old stuff...
        build_package = os.path.join(BUILD_ROOT, 'sandbox')
        if os.path.exists(build_package):
            shutil.rmtree(build_package)
            time.sleep(0.1)

        from neobuilder.neobuilder import NeoBuilder

        # Build stuff...
        builder = NeoBuilder(package='sandbox',
                             protopath=PROTO_ROOT,
                             build_root=BUILD_ROOT)
        builder.build()

        # Add build root to path to access its modules
        sys.path.append(BUILD_ROOT)

    def test_dataclass_to_dict(self):
        from sandbox.test import rainbow_dc

        expected = {'simple_field': 'Green',
                    'message_field': {
                        'foo': 'Blue',
                        'bar': 'Yellow'
                    },
                    'simple_list': ['Four', 'Seven', '99'],
                    'message_list': [{'foo': 'Foo in a list',
                                      'bar': 'Candybar'},
                                     {'foo': 'a Fool in a list',
                                      'bar': 'Tequila bar'}],
                    'simple_map': {'dora': 'Imamap!',
                                   'diego': 'Camera!'},
                    'message_map': {'mickey': {'foo': 'mouse',
                                               'bar': ''},  # TODO(thordurm@ccpgames.com>) 2024-04-15: Should we include "default/empty" values?!?
                                    'donald': {'foo': 'duck',
                                               'bar': 'trump'}}}

        dc = rainbow_dc.RainbowMessage(simple_field='Green',
                                       message_field=rainbow_dc.SubMessage(foo='Blue',
                                                                           bar='Yellow'),
                                       simple_list=['Four', 'Seven', '99'],
                                       message_list=[rainbow_dc.SubMessage(foo='Foo in a list',
                                                                           bar='Candybar'),
                                                     rainbow_dc.SubMessage(foo='a Fool in a list',
                                                                           bar='Tequila bar')],
                                       simple_map={'dora': 'Imamap!',
                                                   'diego': 'Camera!'},
                                       message_map={'donald': rainbow_dc.SubMessage(foo='duck',
                                                                                    bar='trump'),
                                                    'mickey': rainbow_dc.SubMessage(foo='mouse')})

        self.assertEqual(expected, dictator.dataclass_to_dict(dc))

        dc2 = rainbow_dc.RainbowMessage(simple_field='Green',
                                        message_field=rainbow_dc.SubMessage(foo='Blue',
                                                                            bar='Yellow'),
                                        simple_list=['Four', 'Seven', '99'],
                                        message_list=[rainbow_dc.SubMessage(foo='Foo in a list',
                                                                            bar='Candybar'),
                                                      rainbow_dc.SubMessage(foo='a Fool in a list',
                                                                            bar='Tequila bar')],
                                        simple_map={'dora': 'Imamap!',
                                                    'diego': 'Camera!'},
                                        message_map={'donald': rainbow_dc.SubMessage(foo='duck',
                                                                                     bar='trump'),
                                                     'mickey': rainbow_dc.SubMessage(foo='rat')})

        self.assertNotEqual(expected, dictator.dataclass_to_dict(dc2))

    def test_dataclass_timestamp_to_dict(self):
        from sandbox.test import rainbow_dc

        ts1 = '2012-07-03T14:50:51.654321Z'
        dt1 = datetime.datetime(2012, 7, 3, 14, 50, 51, 654321)
        self.assertEqual(ts1, dt1.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts2 = '2010-04-09T22:38:39.009870Z'
        dt2 = datetime.datetime(2010, 4, 9, 22, 38, 39, 9870)
        self.assertEqual(ts2, dt2.strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

        ts3 = '1979-07-06T14:30:00Z'
        dt3 = datetime.datetime(1979, 7, 6, 14, 30)
        self.assertEqual(ts3, dt3.strftime('%Y-%m-%dT%H:%M:%SZ'))

        ts4 = '2049-01-01T12:00:00Z'
        dt4 = datetime.datetime(2049, 1, 1, 12)
        self.assertEqual(ts4, dt4.strftime('%Y-%m-%dT%H:%M:%SZ'))

        ts5 = '2049-01-01T00:00:00Z'
        dt5 = datetime.datetime(2049, 1, 1)
        self.assertEqual(ts5, dt5.strftime('%Y-%m-%dT%H:%M:%SZ'))

        expected = {'my_timestamp': ts1,
                    'my_timestamp_list': [ts2, ts3],
                    'my_timestamp_map': {
                        'noon': ts4,
                        'midnight': ts5}}

        dc = rainbow_dc.TimestampMessage(my_timestamp=dt1,
                                         my_timestamp_list=[dt2, dt3],
                                         my_timestamp_map={'noon': dt4,
                                                           'midnight': dt5})

        self.assertEqual(expected, dictator.dataclass_to_dict(dc))

    def test_dataclass_bytes_to_dict(self):
        from sandbox.test import rainbow_dc

        as_str_1 = 'Þórður Matthíasson'
        as_bytes_1 = as_str_1.encode('utf-8')
        as_base64_1 = base64.encodebytes(as_bytes_1).decode('utf-8').strip()

        as_str_2 = '♡'
        as_bytes_2 = as_str_2.encode('utf-8')
        as_base64_2 = base64.encodebytes(as_bytes_2).decode('utf-8').strip()

        as_str_3 = '♠'
        as_bytes_3 = as_str_3.encode('utf-8')
        as_base64_3 = base64.encodebytes(as_bytes_3).decode('utf-8').strip()

        as_str_4 = '♢'
        as_bytes_4 = as_str_4.encode('utf-8')
        as_base64_4 = base64.encodebytes(as_bytes_4).decode('utf-8').strip()

        as_str_5 = '♣'
        as_bytes_5 = as_str_5.encode('utf-8')
        as_base64_5 = base64.encodebytes(as_bytes_5).decode('utf-8').strip()

        expected = {'my_bytes': as_base64_1,
                    'my_bytes_list': [as_base64_2, as_base64_3],
                    'my_bytes_map': {
                        'zero': as_base64_4,
                        'one': as_base64_5}}

        dc = rainbow_dc.BytesMessage(my_bytes=as_bytes_1,
                                     my_bytes_list=[as_bytes_2, as_bytes_3],
                                     my_bytes_map={'zero': as_bytes_4,
                                                   'one': as_bytes_5})

        self.assertEqual(expected, dictator.dataclass_to_dict(dc))

    def test_dataclass_enums_to_dict(self):
        from sandbox.test import enums_dc
        dc1 = enums_dc.WithExternalEnum()
        dc1.my_enum = enums_dc.TWO

        dc1.my_enum_list.append(enums_dc.ONE)
        dc1.my_enum_list.append(enums_dc.THREE)

        dc1.my_enum_map['one'] = enums_dc.ONE
        dc1.my_enum_map['two'] = enums_dc.TWO

        dc1.my_alias_enum = enums_dc.SIX

        dc1.my_alias_enum_list.append(enums_dc.FOUR)
        dc1.my_alias_enum_list.append(enums_dc.FIVE)
        dc1.my_alias_enum_list.append(enums_dc.FJORIR)

        dc1.my_alias_enum_map['six'] = enums_dc.SIX
        dc1.my_alias_enum_map['sex'] = enums_dc.SEX
        dc1.my_alias_enum_map['fimm'] = enums_dc.FIVE

        expected = {'my_enum': 2,
                    'my_alias_enum': 3,
                    'my_enum_list': [1, 3],
                    'my_alias_enum_list': [1, 2, 1],
                    'my_enum_map': {'one': 1,
                                    'two': 2},
                    'my_alias_enum_map': {'six': 3,
                                          'sex': 3,
                                          'fimm': 2}}

        self.assertEqual(expected, dictator.dataclass_to_dict(dc1))

        dc2 = enums_dc.WithInternalEnum()
        dc2.my_internal_enum = enums_dc.WithInternalEnum.SIX

        dc2.my_internal_enum_list.append(enums_dc.WithInternalEnum.FIVE)
        dc2.my_internal_enum_list.append(enums_dc.WithInternalEnum.FOUR)

        dc2.my_internal_enum_map['no5'] = enums_dc.WithInternalEnum.FIVE
        dc2.my_internal_enum_map['no6'] = enums_dc.WithInternalEnum.SIX

        dc2.my_internal_alias_enum = enums_dc.WithInternalEnum.SEVEN

        dc2.my_internal_alias_enum_list.append(enums_dc.WithInternalEnum.SJO)
        dc2.my_internal_alias_enum_list.append(enums_dc.WithInternalEnum.ATTA)
        dc2.my_internal_alias_enum_list.append(enums_dc.WithInternalEnum.EIGHT)

        dc2.my_internal_alias_enum_map['no9'] = enums_dc.WithInternalEnum.NIU
        dc2.my_internal_alias_enum_map['no9B'] = enums_dc.WithInternalEnum.NINE
        dc2.my_internal_alias_enum_map['default'] = enums_dc.WithInternalEnum.ZERO

        expected2 = {'my_internal_enum': 6,
                     'my_internal_alias_enum': 7,
                     'my_internal_enum_list': [5, 4],
                     'my_internal_alias_enum_list': [7, 8, 8],
                     'my_internal_enum_map': {'no5': 5,
                                              'no6': 6},
                     'my_internal_alias_enum_map': {'no9': 9,
                                                    'no9B': 9,
                                                    'default': 0}}

        self.assertEqual(expected2, dictator.dataclass_to_dict(dc2))

    def test_proto_struct_to_dict(self):
        from sandbox.test.googlestruct_dc import StructMessage
        expected_a = {
            'my_bool': True,
            'my_float': 4.2,
            'my_null': None,
            'my_dict': {'foo': 'bar', 'you': 'tube'},
            'my_list': [1.0, 3.0, 5.0, 8.0],
            'my_string': 'I am String, hear me spell!',
            'my_int': 42.0
        }

        expected_b = {
            'my_struct': expected_a
        }

        struct_dc = StructMessage(my_struct={
            'my_bool': True,
            'my_float': 4.2,
            'my_null': None,
            'my_dict': {'foo': 'bar', 'you': 'tube'},
            'my_list': [1.0, 3.0, 5.0, 8.0],
            'my_string': 'I am String, hear me spell!',
            'my_int': 42.0
        })

        self.assertEqual(expected_b, dictator.dataclass_to_dict(struct_dc))



__all__ = [
    'MethodIoType',
]

import enum


class MethodIoType(enum.IntFlag):
    _UNKNOWN = 0b0000  # 0
    _INPUT_UNARY = 0b0001  # 1
    _INPUT_STREAM = 0b0010  # 2
    _INPUT_MASK = 0b0011  # 3

    _OUTPUT_UNARY = 0b0100  # 4
    _OUTPUT_STREAM = 0b1000  # 8
    _OUTPUT_MASK = 0b1100  # 12

    UNARY_IN_UNARY_OUT = _INPUT_UNARY | _OUTPUT_UNARY    # 0b0101  # 5
    UNARY_IN_STREAM_OUT = _INPUT_UNARY | _OUTPUT_STREAM   # 0b0110  # 6
    STREAM_IN_UNARY_OUT = _INPUT_STREAM | _OUTPUT_UNARY   # 0b1001  # 9
    STREAM_IN_STREAM_OUT = _INPUT_STREAM | _OUTPUT_STREAM  # 0b1010  # 10

    def is_valid(self) -> bool:
        return self.value in (self.UNARY_IN_UNARY_OUT.value,
                              self.UNARY_IN_STREAM_OUT.value,
                              self.STREAM_IN_UNARY_OUT.value,
                              self.STREAM_IN_STREAM_OUT.value)

    def is_input_stream(self) -> bool:
        return (self.value & self._INPUT_STREAM) == self._INPUT_STREAM

    def is_output_stream(self) -> bool:
        return (self.value & self._OUTPUT_STREAM) == self._OUTPUT_STREAM



__all__ = [
    'force_object',
    'kwdict',
    'import_dataclass_by_proto',
    'get_type_from_module_import_as',
    'get_dc_field_obj_type',
    'humps_to_under',
    'check_type',
    'get_type_url',
    'get_type_name_from_url',
    'get_proto_descriptor_by_type_url',
    'get_proto_file_by_type_url',
    'base64_stripper',
    'base64_filler',
    'fuzzy_base64_to_bytes',
]

from collections import abc
import sys
import re
import builtins
import base64
import typing
import enum

from google.protobuf import symbol_database
from google.protobuf import json_format
from ccptools.tpu import strimp

from protoplasm.structs import *

import logging
log = logging.getLogger(__name__)


T_ANY_CLASS = TypeVar('T_ANY_CLASS')

TYPING_MATCHER = re.compile(r'typing.(?P<type>[a-zA-Z]+)\[(?P<inner>[a-zA-Z0-9.,_ ]+)]')
STAR_IMPORT_TYPING_MATCHER = re.compile(r'(?P<type>List|Dict)\[(?P<inner>.+)]')

_POST_39 = sys.version_info >= (3, 9)


def force_object(object_cls: Type[T_ANY_CLASS], value: Any) -> T_ANY_CLASS:
    """Takes a `value` in the form of a single argument, list of arguments, a
    dict of keyword arguments or an object of type `object_cls`.

    If the value is already an object of type `object_cls` or a `None`, it's
    simply returned as is.

    If the `value` is a single argument variable, it is used as the first (and
    only) arg to instantiate a new object of the `object_cls` type which is then
    returned.

    If the `value` is a `list` or `tuple`, the values therein are used as
    `*args` to instantiate a new object of the `object_cls` type which is then
    returned.

    If the `value` is a `dict`, the keys and values therein are used as
    `**kwargs` to  instantiate a new object of the `object_cls` type which is
    then returned.

    This guarantees a return value of either an object of `object_cls` type, a
    none or a raised `Exception`, e.g. if the supplied `value` can't be used as
    `__init__()` arguments in the `object_cls` (most likely a `TypeError` or
    `ValueError`) or if the `__init__()` of that class raises any other
    exception.

    If you need a `None` value forced into a new empty instance of the given
    `object_cls` a simple way to achieve that is:

    ```my_foo = force_object(Foo, my_values) or Foo()```

    One thing worth mentioning is that if `object_cls` takes a `list`, `tuple`
    or a `dict` as it's first `__init__()` argument and the given `value` is of
    that same type and intended to be used as the first and only argument for
    the `__init__()`, that will not work as `value` of those three types will be
    treated as a collection of `*args` or mapping og `**kwargs` for the
    `__init__()` instead of a single argument.

    Solutions to that include:
     - Not using this function for classes that take `list`, `tuple` or `dict`
       as their first arguments. This isn't intended to be a universal solution
       that fits all cases and classes so if you have a square hole, don't use a
       round peg. ;)
     - Wrapping the `value` in a `list` or `tuple` of its own before passing it
       to this function as that will get treated as a list of `*args` for the
       `__init__()`, with the inner `list`, `tuple` or `dict` used correctly as
       the first (and only) argument.
        - This may however not work correctly if `value` is a None and the
          `__init__()` doesn't take kindly to None values for the argument it
          expects to be a `list`, `tuple` or `dict`

    :param object_cls: The class/type to force the given value into
    :param value:
    :type value:
    :return:
    :rtype:
    """
    if value is None or isinstance(value, object_cls):
        return value
    if isinstance(value, dict):
        return object_cls(**value)
    elif isinstance(value, (list, tuple)):
        return object_cls(*value)
    else:
        return object_cls(value)


def kwdict(**kwargs) -> dict:
    """Takes kwargs and turns into a dict using double underscores in order to
    build nested dicts.

    Examples:

    >>> kwdict(foo='one', bar='two')
    {'foo': 'one', 'bar': 'two'}

    >>> kwdict(foo='one', bar__one=1, bar__two=2)
    {'foo': 'one', 'bar': {'one': 1, 'two': 2}}
    """
    d = {}
    subs = {}
    for k, v in kwargs.items():
        first, *rest = k.split('__', 1)
        if rest:
            if first not in subs:
                subs[first] = {}
            subs[first][rest[0]] = v
        else:
            d[first] = v

    for k, v in subs.items():
        d[k] = kwdict(**v)

    return d


def import_dataclass_by_proto(proto_class: GeneratedProtocolMessageType) -> Type[DataclassBase]:
    """Dynamically imports and returns the dataclass type corresponding to the
    given Protobuf message class.

    This will also cache that dataclass type in the Protobuf message class in
    order to speed up subsequent future imports by a tiny tiny fraction.

    :param proto_class: The Protobuf message class that we need to find the
                        dataclass version of.
    """
    cls = getattr(proto_class, '__dataclass_cls__', None)
    if not cls:
        if not isinstance(proto_class, type):
            proto_class = proto_class.__class__
        cls = strimp.get_class(''.join([proto_class.__module__[:-3], 'dc.', proto_class.__name__]), reraise=True)
        setattr(proto_class, '__dataclass_cls__', cls)
    return cls


# TODO(thordurm@ccpgames.com>): Maybe this should live in typeutils strimp?!?
def get_type_from_module_import_as(type_name: str, cls: Type) -> Type:
    cls_module = sys.modules.get(cls.__module__)
    if not cls_module:
        raise ImportError(f'get_type_from_module_import_as failed to find {cls.__module__} in sys.modules')

    if '.' in type_name:
        m = TYPING_MATCHER.match(type_name)
        if m:  # From 'typing' (this is pre-version 4.1 stuff, kept in for backwards compatibility)
            return _get_from_typing(m.group('type'), m.group('inner'), cls)
        else:
            m = STAR_IMPORT_TYPING_MATCHER.match(type_name)
            if m:  # From 'typing' via 'from typing import *'
                return _get_from_typing(m.group('type'), m.group('inner'), cls)

            else:  # From somewhere else
                as_name, type_name = type_name.rsplit('.', 1)
                cls_module = sys.modules.get(cls.__module__)
                if not cls_module:
                    raise ImportError(f'get_type_from_module_import_as failed to find {cls.__module__} in sys.modules')

                alias_module = cls_module
                for as_part in as_name.split('.'):
                    alias_module = getattr(alias_module, as_part, None)

                    if not alias_module:
                        raise ImportError(f'get_type_from_module_import_as failed to find {as_name} in {cls_module}')
    else:
        builtin_type = getattr(builtins, type_name, None)
        if not builtin_type:
            m = STAR_IMPORT_TYPING_MATCHER.match(type_name)
            if m:  # From 'typing' via 'from typing import *'
                return _get_from_typing(m.group('type'), m.group('inner'), cls)
        else:
            if isinstance(builtin_type, type):
                return builtin_type

        alias_module = cls_module

    actual_type = getattr(alias_module, type_name, None)
    if not actual_type:
        raise ImportError(f'get_type_from_module_import_as failed to find {actual_type} in {alias_module}')

    return actual_type


def _get_from_typing(type_name: str, inner_str: str, cls: Type) -> Type:
    typing_type = getattr(typing, type_name, None)  # E.g. List
    if not typing_type:
        raise ImportError(f'get_type_from_module_import_as->_get_from_typing failed to find {type_name} in typing module')
    if inner_str:
        args = tuple([get_type_from_module_import_as(a.strip(), cls) for a in inner_str.split(',')])
        return typing_type[args]
    return typing_type


def get_dc_field_obj_type(dc_field: dataclasses.Field,
                          parent_dataclass: Type[DataclassBase]) -> Union[Type, Type[DataclassBase], enum.EnumMeta]:
    """Fetches the "actual" type of a dataclass field. In the case of lists and
    maps/dicts this means the type of the values the lists/maps/dicts contain.

    This assumes a dataclass generated by protoplasm.

    :param dc_field: The dataclass field who's type we need.
    :param parent_dataclass: The dataclass itself
    :return: The type defined in the dataclass field
    """
    if isinstance(dc_field.type, str):
        dc_field.type = get_type_from_module_import_as(dc_field.type, parent_dataclass)

    if dc_field.metadata.get('is_map', False):
        return dc_field.type.__args__[1]
    elif dc_field.metadata.get('is_list', False):
        return dc_field.type.__args__[0]
    else:
        return dc_field.type


def humps_to_under(string: str) -> str:
    buf = []
    if string.endswith('ID'):
        if len(string) > 2 and string[-3].islower():
            string = f'{string[:-2]}_id'
        else:
            return string.lower()

    for i, c in enumerate(string):
        if c.isupper():
            if i != 0:
                buf.append('_')
            c = c.lower()
        buf.append(c)
    return ''.join(buf)


def check_type(val: Any, type_or_annotation: Any) -> bool:
    # TODO(thordurm@ccpgames.com>): Move to or copy to typeutils

    def _is_special(toa: Any) -> bool:
        if _POST_39:
            if isinstance(toa, typing._SpecialGenericAlias):  # noqa
                return True
        else:
            return toa._special  # noqa

    if val is None and type_or_annotation is None:
        return True

    m = getattr(type_or_annotation, '__module__', None)
    if m == 'typing':
        if type_or_annotation == Any:
            return True
        if isinstance(type_or_annotation, typing._GenericAlias):  # noqa
            if _is_special(type_or_annotation):  # noqa Not subscripted
                return isinstance(val, type_or_annotation.__origin__)

            else:
                if type_or_annotation.__origin__ == Union:
                    if not type_or_annotation.__args__:
                        return True
                    if Any in type_or_annotation.__args__:
                        return True
                    return isinstance(val, type_or_annotation.__args__)

                elif type_or_annotation.__origin__ in (list, set):  # Check list and set
                    if not isinstance(val, type_or_annotation.__origin__):
                        return False
                    if not type_or_annotation.__args__ or Any in type_or_annotation.__args__:
                        return True
                    for sub_val in val:
                        if not check_type(sub_val, Union[type_or_annotation.__args__]):
                            return False
                    return True  # Should be good now! :)

                elif type_or_annotation.__origin__ == tuple:  # Check tuple!
                    if not isinstance(val, tuple):
                        return False
                    if not type_or_annotation.__args__:
                        return True

                    if type_or_annotation.__args__[-1] is ... and len(type_or_annotation.__args__) == 2:
                        for sub_val in val:
                            if not check_type(sub_val, Union[type_or_annotation.__args__[0]]):
                                return False
                        return True  # Should be good now! :)

                    if len(type_or_annotation.__args__) != len(val):
                        return False

                    for i, sub_val in enumerate(val):
                        if not check_type(sub_val, Union[type_or_annotation.__args__[i]]):
                            return False

                    return True  # Should be good now! :)

                elif issubclass(type_or_annotation.__origin__, abc.Mapping):
                    if not isinstance(val, type_or_annotation.__origin__):
                        return False
                    if not type_or_annotation.__args__ or type_or_annotation.__args__ == (Any, Any):
                        return True

                    for k, v in val.items():
                        if not (check_type(k, Union[type_or_annotation.__args__[0]]) and check_type(v, Union[type_or_annotation.__args__[1]])):
                            return False
                    return True
                else:
                    log.warning('I do not know how to check type %r of typing module:(', type_or_annotation)
        if isinstance(type_or_annotation, TypeVar):
            # TODO(thordurm@ccpgames.com>): Not really tested and supported yet officially...
            # TODO(thordurm@ccpgames.com>): Main issue is nested types (if they don't flatten)
            return check_type(val, Union[type_or_annotation.__constraints__])
    return isinstance(val, type_or_annotation)


def get_type_url(proto: GeneratedProtocolMessageType, prefix: str = 'type.googleapis.com/') -> str:
    # TODO(thordurm@ccpgames.com>) 2024-04-15: Not sure how best to support any prefix other than "type.googleapis.com/" here
    if not prefix:
        return f'/{proto.DESCRIPTOR.full_name}'
    if prefix[-1] == '/':
        return f'{prefix}{proto.DESCRIPTOR.full_name}'
    return f'{prefix}/{proto.DESCRIPTOR.full_name}'


def get_type_name_from_url(type_url: str) -> str:
    # TODO(thordurm@ccpgames.com>) 2024-04-15: Not sure how best to support any prefix other than "type.googleapis.com/" here
    return type_url.split('/')[-1]


def get_proto_descriptor_by_type_url(type_url: str) -> json_format.descriptor.Descriptor:
    # TODO(thordurm@ccpgames.com>) 2024-04-15: Not sure how best to support any prefix other than "type.googleapis.com/" here
    type_name = get_type_name_from_url(type_url)
    db = symbol_database.Default()
    try:
        message_descriptor = db.pool.FindMessageTypeByName(type_name)
    except KeyError:
        raise TypeError('Can not find message descriptor by type_url: {0}. You may need to import the pb2 file'
                        ' before it registers in the SymbolDatabase'.format(type_url))
    return message_descriptor


def get_proto_file_by_type_url(type_url: str) -> json_format.descriptor.FileDescriptor:
    # TODO(thordurm@ccpgames.com>) 2024-04-15: Not sure how best to support any prefix other than "type.googleapis.com/" here
    type_name = get_type_name_from_url(type_url)
    db = symbol_database.Default()
    try:
        file_descriptor = db.pool.FindFileContainingSymbol(type_name)
    except KeyError:
        raise TypeError('Can not find file descriptor by type_url: {0}. You may need to import the pb2 file before'
                        ' it registers in the SymbolDatabase'.format(type_url))
    return file_descriptor


def base64_stripper(base64str: Union[bytes, str]) -> bytes:
    if isinstance(base64str, str):
        base64str = base64str.encode()
    return base64str.strip().replace(b'\r', b'').replace(b'\n', b'').replace(b'\t', b'').replace(b'=', b'')


def base64_filler(base64str: Union[bytes, str]) -> bytes:
    if isinstance(base64str, str):
        base64str = base64str.encode()
    base64str = base64str.strip().replace(b'\r', b'').replace(b'\n', b'').replace(b'\t', b'')
    return base64str + b'=' * (4 - (len(base64str) % 4) or 4)


def fuzzy_base64_to_bytes(base64str: Union[bytes, str]) -> bytes:
    return base64.decodebytes(base64_filler(base64str))



__all__ = [
    'mkproto',
    'mkdataclass',
    'proto_to_dataclass',
    'dataclass_to_proto',
    'unpack_dataclass_return',
    'pack_dataclass_response',
]
# Alias imports
from protoplasm.structs import *
from protoplasm.casting.castutils import *
from protoplasm.casting.dictator import *
from protoplasm.casting.objectifier import *

import logging
log = logging.getLogger(__name__)


def mkproto(proto_class: Type[GeneratedProtocolMessageType], **kwargs) -> GeneratedProtocolMessageType:
    """Creates a new Protobuf object of the given proto_class type and populates
    it using the given kwargs, first converting the kwargs to a dict via the
    kwdict function that turns keywords using double-underscores into sub-dicts
    for use as data for nested proto messages.

    :param proto_class: The Protobuf message class (type)
    :return: A fully initialized and populated Profobuf object
    """
    return dict_to_proto(proto_class, kwdict(**kwargs))


def mkdataclass(dataclass_type: Type[T_DCB], **kwargs) -> T_DCB:
    """Creates a new Dataclass object of the given dataclass_type and populates
    it using the given kwargs, first converting the kwargs to a dict via the
    kwdict function that turns keywords using double-underscores into sub-dicts
    for use as data for nested proto messages.

    :param dataclass_type: Dataclass type to create
    :return: A fully initialized and populated Profobuf object
    """
    return dict_to_dataclass(dataclass_type, kwdict(**kwargs))


def proto_to_dataclass(proto: GeneratedProtocolMessageType) -> DataclassBase:
    return dict_to_dataclass(import_dataclass_by_proto(proto.__class__), proto_to_dict(proto))


def dataclass_to_proto(dc: DataclassBase) -> GeneratedProtocolMessageType:
    return dict_to_proto(dc.__proto_cls__, dataclass_to_dict(dc))


def unpack_dataclass_return(dc: DataclassBase) -> Optional[Tuple[Any]]:
    l = []
    for field in dataclasses.fields(dc):
        l.append(getattr(dc, field.name))
    ln = len(l)
    if ln == 0:
        return
    elif ln == 1:
        return l[0]
    return tuple(l)


def pack_dataclass_response(dc_response_type: Type[T_DCB], return_value) -> T_DCB:
    if isinstance(return_value, dc_response_type):
        return return_value

    if return_value is None:
        return dc_response_type()

    if isinstance(return_value, dict):
        return dc_response_type(**return_value)  # TODO(thordurm@ccpgames.com>): Temporary solution
        # TODO(thordurm@ccpgames.com>): If keys match dc_response_type fields, it's a simple kwargs mapping
        # TODO(thordurm@ccpgames.com>): ...else If the first field it a dict, treat this as the first field
        # TODO(thordurm@ccpgames.com>): ...else we're fucked!

    elif isinstance(return_value, tuple):
        return dc_response_type(*return_value)  # TODO(thordurm@ccpgames.com>): Temporary solution
        # TODO(thordurm@ccpgames.com>): If value types match field types and len is <= fields, it's a simple args mapping
    else:
        return dc_response_type(return_value)



__all__ = [
    'GrpcContext',
]

import re
import grpc
from ._basectx import *
from protoplasm.structs._methodtype import *


class GrpcContext(BaseContext):
    IP_MATCHER = re.compile(r'(?:(?P<ipv>ipv[46]):)?(?P<ip>(?:\[?[\da-f]{0,4}(?::[\da-f]{0,4}){2,8}]?)|(?:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}))(?P<port>:\d{1,5})?', re.IGNORECASE)

    def __init__(self, wrapped_context: grpc.ServicerContext, method_io: MethodIoType = None):
        super().__init__(wrapped_context)
        self.metadata = {}
        self.method_io = method_io or MethodIoType._UNKNOWN  # noqa
        self._load_metadata()
        self._origin_ip = None

    def _load_metadata(self):
        self.metadata = {}
        for k, v in self.wrapped_context.invocation_metadata():
            k = k.lower()
            if k not in self.metadata:
                self.metadata[k] = v
            else:
                if not isinstance(self.metadata[k], list):
                    self.metadata[k] = [self.metadata[k]]
                self.metadata[k].append(v)

    def get_origin_ip(self, default=None):
        if not self._origin_ip:
            if self._origin_ip is False:
                return default
            # Try x-forwarded!
            self._origin_ip = self.get_meta('x-forwarded-for', None)
            if not self._origin_ip:
                # If not, try peer
                peer = self.wrapped_context.peer()
                if peer:
                    m = GrpcContext.IP_MATCHER.match(peer)
                    if m:
                        self._origin_ip = m.group('ip')
            if not self._origin_ip:
                self._origin_ip = False
        return self._origin_ip

    def get_meta(self, name, default=None):
        return self.metadata.get(name, default)

    def __repr__(self):
        return 'GrpcContext[metadata=%r, peer=%r]' % (self.metadata, self.wrapped_context.peer())



import unittest
from protoplasm import casting
import os
import sys

import shutil
import time
HERE = os.path.dirname(__file__)
PROTO_ROOT = os.path.join(HERE, 'res', 'proto')
BUILD_ROOT = os.path.join(HERE, 'res', 'build')

import logging
log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


def _get_sub():
    from sandbox.test import rainbow_dc
    return rainbow_dc.SubMessage.from_dict({'foo': 'Foo!', 'bar': 'Bar!'})


class DataclassTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Remove old stuff...
        build_package = os.path.join(BUILD_ROOT, 'sandbox')
        if os.path.exists(build_package):
            shutil.rmtree(build_package)
            time.sleep(0.1)

        from neobuilder.neobuilder import NeoBuilder

        # Build stuff...
        builder = NeoBuilder(package='sandbox',
                             protopath=PROTO_ROOT,
                             build_root=BUILD_ROOT)
        builder.build()

        # Add build root to path to access its modules
        sys.path.append(BUILD_ROOT)

    def test_args(self):
        from sandbox.test import rainbow_dc

        dc1 = rainbow_dc.RainbowMessage()
        dc1.simple_field = 'I iz string'
        dc1.message_field = rainbow_dc.SubMessage.from_dict({'foo': 'Foo!', 'bar': 'Bar!'})
        dc1.simple_list = ['one', 'two', "Freddy's coming for you!"]
        dc1.message_list = [rainbow_dc.SubMessage(foo='Foo1!', bar='Bar1!'),
                            rainbow_dc.SubMessage(foo='Foo2!', bar='Bar2!'),
                            rainbow_dc.SubMessage(foo='Foo3!', bar='Bar3!')]
        dc1.simple_map = {'uno': 'einn', 'dos': 'tveir'}
        dc1.message_map = {'ein': rainbow_dc.SubMessage(foo='Foo11!', bar='Bar11!'),
                           'zwei': rainbow_dc.SubMessage(foo='Foo22!', bar='Bar22!')}

        dc2 = rainbow_dc.RainbowMessage('I iz string',
                                        rainbow_dc.SubMessage(foo='Foo!', bar='Bar!'),
                                        ['one', 'two', "Freddy's coming for you!"],
                                        [rainbow_dc.SubMessage(foo='Foo1!', bar='Bar1!'),
                                         rainbow_dc.SubMessage(foo='Foo2!', bar='Bar2!'),
                                         rainbow_dc.SubMessage(foo='Foo3!', bar='Bar3!')],
                                        {'uno': 'einn', 'dos': 'tveir'},
                                        {'ein': rainbow_dc.SubMessage(foo='Foo11!', bar='Bar11!'),
                                         'zwei': rainbow_dc.SubMessage(foo='Foo22!', bar='Bar22!')})

        self.assertEqual(dc1, dc2)

    def test_shortcut_args(self):
        from sandbox.test import rainbow_dc

        dc1 = rainbow_dc.RainbowMessage('I iz string', 'Foo!')
        dc2 = rainbow_dc.RainbowMessage('I iz string', rainbow_dc.SubMessage(foo='Foo!'))

        dc2b = rainbow_dc.RainbowMessage.from_kwdict(simple_field='I iz string', message_field='Foo!')

        self.assertEqual(dc1, dc2)
        self.assertEqual(dc1, dc2b)

        dc3 = rainbow_dc.RainbowMessage('I iz string', ('Foo!', 'Bar!'))
        dc4 = rainbow_dc.RainbowMessage('I iz string', {'foo': 'Foo!', 'bar': 'Bar!'})
        dc5 = rainbow_dc.RainbowMessage('I iz string', rainbow_dc.SubMessage(foo='Foo!', bar='Bar!'))
        dc5b = rainbow_dc.RainbowMessage.from_kwdict(simple_field='I iz string', message_field=('Foo!', 'Bar!'))

        self.assertEqual(dc3, dc4)
        self.assertEqual(dc3, dc5)
        self.assertEqual(dc3, dc5b)
        self.assertNotEqual(dc1, dc3)

    def test_shortcut_list_args(self):
        from sandbox.test import rainbow_dc
        dc1 = rainbow_dc.RainbowMessage(message_list=['Foo!', 'Foo2!'])
        dc2 = rainbow_dc.RainbowMessage(message_list=[rainbow_dc.SubMessage(foo='Foo!'),
                                                      rainbow_dc.SubMessage(foo='Foo2!')])

        self.assertEqual(dc1, dc2)

        dc3 = rainbow_dc.RainbowMessage(message_list=[('Foo!', 'Bar!'), ('Foo2!', 'Bar2!')])
        dc4 = rainbow_dc.RainbowMessage(message_list=[{'foo': 'Foo!', 'bar': 'Bar!'}, {'foo': 'Foo2!', 'bar': 'Bar2!'}])
        dc5 = rainbow_dc.RainbowMessage(message_list=[rainbow_dc.SubMessage(foo='Foo!', bar='Bar!'),
                                                      rainbow_dc.SubMessage(foo='Foo2!', bar='Bar2!')])
        dc6 = rainbow_dc.RainbowMessage(message_list=[('Foo!', 'Bar!'), ('Foo2!', 'Bar2!')])
        dc7 = rainbow_dc.RainbowMessage(message_list=[('Foo!', 'Bar!'), {'foo': 'Foo2!', 'bar': 'Bar2!'}])
        dc8 = rainbow_dc.RainbowMessage(message_list=[rainbow_dc.SubMessage(foo='Foo!', bar='Bar!'), {'foo': 'Foo2!', 'bar': 'Bar2!'}])
        self.assertEqual(dc3, dc4)
        self.assertEqual(dc3, dc5)
        self.assertEqual(dc3, dc6)
        self.assertEqual(dc3, dc7)
        self.assertEqual(dc3, dc8)
        self.assertNotEqual(dc1, dc3)

    def test_shortcut_dict_args(self):
        from sandbox.test import rainbow_dc
        dc1 = rainbow_dc.RainbowMessage(message_map={'one': 'Foo!', 'two': 'Foo2!'})
        dc2 = rainbow_dc.RainbowMessage(message_map={'one': rainbow_dc.SubMessage(foo='Foo!'),
                                                     'two': rainbow_dc.SubMessage(foo='Foo2!')})
        dc2b = rainbow_dc.RainbowMessage.from_kwdict(message_map={'one': 'Foo!', 'two': 'Foo2!'})

        self.assertEqual(dc1, dc2)
        self.assertEqual(dc1, dc2b)

        dc3 = rainbow_dc.RainbowMessage(message_map={'one': ('Foo!', 'Bar!'), 'two': ('Foo2!', 'Bar2!')})
        dc4 = rainbow_dc.RainbowMessage(message_map={'one': {'foo': 'Foo!', 'bar': 'Bar!'}, 'two': {'foo': 'Foo2!', 'bar': 'Bar2!'}})
        dc5 = rainbow_dc.RainbowMessage(message_map={'one': rainbow_dc.SubMessage(foo='Foo!', bar='Bar!'),
                                                     'two': rainbow_dc.SubMessage(foo='Foo2!', bar='Bar2!')})
        self.assertEqual(dc3, dc4)
        self.assertEqual(dc3, dc5)
        self.assertNotEqual(dc1, dc3)



__all__ = [
    'RemoteMethodWrapper',
]
import json
import uuid

import grpc  # noqa
from grpc._channel import _UnaryUnaryMultiCallable as _UnaryUnaryStub  # noqa
from grpc._channel import _UnaryStreamMultiCallable as _UnaryStreamStub  # noqa
from grpc._channel import _StreamUnaryMultiCallable as _StreamUnaryStub  # noqa
from grpc._channel import _StreamStreamMultiCallable as _StreamStreamStub  # noqa

from ccptools.tpu import casting as typeutils_casting

from protoplasm.structs import *
from protoplasm import casting
from protoplasm import errors

from .dataclass_bases import *
from ._exwrap import *

import logging
log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .grpc_bases import BaseGrpcClientImplementation
    from grpc._channel import _MultiThreadedRendezvous  # noqa


class RemoteMethodWrapper:
    JSON_SERIALIZER = typeutils_casting.JsonSafeSerializer()

    __slots__ = ('client', 'request_cls', 'stub_method', 'args', 'kwargs',
                 'request_iterator', 'response_iterator', 'request_dc',
                 'request_proto', 'response_proto', 'response_dc',
                 'unpacked_results', 'method_io', '_ref_code')

    def __init__(self, client: 'BaseGrpcClientImplementation', request_cls: Type[T_DCB],
                 stub_method: Callable, *args, **kwargs):
        """Container and worker for processing BaseGrpcClient (i.e. Client Side)
        requests and responses, packing, unpacking, casting and so on as well as
        handling errors and log-dumping and all that bulky noisy jass that just
        clutters up the actual client code.
        """
        self.client: 'BaseGrpcClientImplementation' = client
        self.request_cls: Type[T_DCB] = request_cls
        self.stub_method: Callable = stub_method
        self.args: List[Any] = args or []
        self.kwargs: Dict[str, Any] = kwargs or {}

        self.request_iterator: Optional[Iterable[DataclassBase]] = None
        self.response_iterator: Optional['_MultiThreadedRendezvous'] = None

        self.request_dc: Optional[DataclassBase] = None
        self.request_proto: Optional[GeneratedProtocolMessageType] = None
        self.response_proto: Optional[GeneratedProtocolMessageType] = None
        self.response_dc: Optional[DataclassBase] = None
        self.unpacked_results: Optional[Any, Tuple[Any]] = None

        self.method_io: MethodIoType = MethodIoType._UNKNOWN  # noqa

        self._ref_code: str = ''
        self._eval_method_io()

    def _eval_method_io(self):
        if isinstance(self.stub_method, _UnaryUnaryStub):
            self.method_io = MethodIoType.UNARY_IN_UNARY_OUT

        elif isinstance(self.stub_method, _UnaryStreamStub):
            self.method_io = MethodIoType.UNARY_IN_STREAM_OUT

        elif isinstance(self.stub_method, _StreamUnaryStub):
            self.method_io = MethodIoType.STREAM_IN_UNARY_OUT

        elif isinstance(self.stub_method, _StreamStreamStub):
            self.method_io = MethodIoType.STREAM_IN_STREAM_OUT

        else:
            self._abort_preflight('Unable to determine stub method io channels (unary vs. stream)')

    @property
    def ref_code(self) -> str:
        if not self._ref_code:
            self._ref_code = str(uuid.uuid4())
        return self._ref_code

    @property
    def method_name(self) -> str:
        m = getattr(self.stub_method, '_method', str(self.stub_method))
        if isinstance(m, bytes):
            return m.decode()
        return m

    def _get_dump(self, **kwargs) -> dict:
        d = {
            'ref_code': self.ref_code,
            'class': str(self.client.__class__),
            'request_cls': str(self.request_cls),
            'method_name': self.method_name,
            'method_io': self.method_io
        }
        if self.args:
            d['args'] = repr(self.args)
        if self.kwargs:
            d['kwargs'] = repr(self.kwargs)

        if self.request_dc:
            d['request_dc'] = repr(self.request_dc)
        if self.request_proto:
            d['request_proto'] = repr(self.request_proto)

        if self.response_proto:
            d['response_proto'] = repr(self.response_proto)

        if self.response_dc:
            d['response_dc'] = repr(self.response_dc)

        if self.unpacked_results:
            d['unpacked_results'] = repr(self.unpacked_results)

        if kwargs:
            d.update(kwargs)

        return d

    def _get_json_dump(self, **kwargs) -> str:
        return json.dumps(self.JSON_SERIALIZER.serialize(self._get_dump(**kwargs)), indent=4)

    def _abort_preflight(self, message: str, ex: Optional[Exception] = None):
        if ex:
            log.exception(f'{message}::{ex!r}\n---\n{self._get_json_dump(details=repr(ex))}')
            raise_msg = f'{message}: {ex!r}'
        else:
            log.exception(f'{message}\n---\n{self._get_json_dump()}')
            raise_msg = message
        raise errors.api.PreRequestError(raise_msg, self.ref_code, True)

    def _abort_inflight(self, message: str, ex: grpc.RpcError):
        gex = GrpcExceptionWrapper(ex)
        reex = gex.get_reraise()
        if reex.should_log:
            d = dict(details=gex.details,
                     debug_message=gex.debug_error_string,
                     trailing_metadata=gex.trailing_metadata,
                     initial_metadata=gex.initial_metadata,
                     gex=gex,
                     status_code=gex.code,
                     ref_code=reex.ref_code)
            log.exception(f'{message}::{ex!r}\n---\n{self._get_json_dump(**d)}')
        raise reex

    def _abort_postflight(self, message: str, ex: Exception):
        log.exception(f'{message}::{ex!r}\n---\n{self._get_json_dump(details=repr(ex))}')
        raise errors.api.PostRequestError(f'{message}: {ex!r}', self.ref_code, True)

    def _abort(self, message: str, ex: Exception):
        log.exception(f'{message}::{ex!r}\n---\n{self._get_json_dump(details=repr(ex))}')
        raise errors.api.Unknown(f'{message}: {ex!r}', self.ref_code, True)

    def prepare(self):
        if not self.method_io.is_input_stream():
            self._mkdc()
            self._d2p()
        else:
            self._find_request_iterator()
            if isinstance(self.request_iterator, (list, tuple)):
                self.request_iterator = RequestIterator(self.request_iterator)

            elif isinstance(self.request_iterator, self.request_cls):
                self.request_iterator = RequestIterator(self.request_iterator)

    def _find_request_iterator(self):
        if not self.kwargs and self.args and len(self.args) == 1:
            self.request_iterator = self.args[0]
            return

        if not self.args and self.kwargs and len(self.kwargs) == 1 and 'request_iterator' in self.kwargs:
            self.request_iterator = self.kwargs['request_iterator']
            return

        self._abort_preflight('Expected a single arg or kwarg called "request_iterator" as input')

    @property
    def request_proto_iterator(self) -> Iterable[GeneratedProtocolMessageType]:
        for raw_req in self.request_iterator:
            self.request_dc = raw_req
            self._d2p()
            yield self.request_proto
        log.debug('Request Iterator Cancelled!')

    def _mkdc(self):
        try:
            self.request_dc = self.request_cls(*self.args, **self.kwargs)
        except Exception as ex:
            self._abort_preflight('Error while populating request class with arguments', ex)

    def _d2p(self):
        try:
            self.request_proto = self.request_dc.to_proto()
        except Exception as ex:
            self._abort_preflight('Error while casting request dataclass to proto', ex)

    def call(self) -> Optional[Union[ResponseIterator[T_DCB], Any, Tuple[Any]]]:
        # Prepare the call
        self.prepare()

        try:
            if self.method_io.is_input_stream():
                if self.method_io.is_output_stream():
                    self.response_iterator = self.stub_method(self.request_proto_iterator)
                else:
                    self.response_proto = self.stub_method(self.request_proto_iterator)

            else:
                if self.method_io.is_output_stream():
                    self.response_iterator = self.stub_method(self.request_proto)
                else:
                    self.response_proto = self.stub_method(self.request_proto)

        except grpc.RpcError as ex:
            self._abort_inflight('RpcError exception while performing gRPC request', ex)

        except Exception as ex:
            self._abort('Unknown exception while performing gRPC request', ex)

        if self.method_io.is_output_stream():
            return ResponseIterator(self)

        else:
            self._p2d()
            self._unpack()
            return self.unpacked_results

    @property
    def iterable_results(self) -> Iterable[T_DCB]:
        for raw_response in self.response_iterator:
            self.response_proto = raw_response
            self._p2d()
            yield self.response_dc

    def close(self):
        if self.response_iterator:
            self.response_iterator.cancel()

    @property
    def is_closed(self) -> Optional[bool]:
        if self.response_iterator:
            return self.response_iterator.cancelled()
        return None  # If we don't have a result stream this doesn't make sense

    def _p2d(self):
        try:
            self.response_dc = casting.proto_to_dataclass(self.response_proto)
        except Exception as ex:
            self._abort_postflight('Error while casting response proto to dataclass', ex)

    def _unpack(self):
        try:
            self.unpacked_results = self.response_dc.explode()
        except Exception as ex:
            self._abort_postflight('Error while unpacking dataclass arguments', ex)



# Alias import for use of Protoplasm.... just go "from protoplasm.plasm import *" or "from protoplasm import plasm"
from protoplasm.structs import *
from protoplasm.errors._base import *  # noqa
from protoplasm.errors._api import *  # noqa
from protoplasm.bases import *
from protoplasm.grpcserver import *
from protoplasm.decorators import *



__all__ = [
    'require_params',
    'takes_context',
]

import functools
import inspect
import typing
from protoplasm import errors
from protoplasm.casting import castutils

import logging
log = logging.getLogger(__name__)


def _map_args_to_params(param_map, arg_list, kw_map):
    mapped_params = {k: None for k in param_map.keys()}
    if arg_list:
        mapped_params.update(dict(zip(list(param_map.keys())[:len(arg_list)], arg_list)))
    if kw_map:
        mapped_params.update(kw_map)
    return mapped_params


def require_params(arg_name: typing.Union[str, typing.Iterable] = '*', *more_arg_names, typecheck=False):
    arg_is_func = False
    func = None
    if isinstance(arg_name, typing.Callable):
        arg_is_func = True
        func = arg_name
        arg_name = '*'

    def wrap(f):
        param_signature: typing.Dict[str, inspect.Parameter] = dict(inspect.signature(f).parameters)
        required_params = set()

        if arg_name == '*':
            key_list = list(param_signature.keys())
            if key_list and key_list[0] in ('self', 'cls'):
                key_list.pop(0)
            ctx_arg = getattr(f, '__takes_context_as__', None)
            if ctx_arg and key_list[-1] == ctx_arg:
                key_list.pop(-1)
            required_params = set(key_list)
        else:
            arg_list = []
            if isinstance(arg_name, (list, tuple, set)):
                arg_list.extend(arg_name)
            else:
                arg_list.append(arg_name)
            if more_arg_names:
                arg_list.extend(more_arg_names)

            arg_list = set(arg_list)

            for a in arg_list:
                if a not in param_signature:
                    raise NameError(f'Required argument {a} not found in function {f!r} -> {param_signature!r}')
                required_params.add(a)

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            given_params = _map_args_to_params(param_signature, args, kwargs)
            for p in required_params:
                val = given_params[p]
                if val is None:
                    raise errors.api.InvalidArgument(f'required parameter "{p}" missing [{f!r}]')
                else:
                    if typecheck:
                        if not castutils.check_type(val, param_signature[p].annotation):
                            raise errors.api.InvalidArgument(f'required parameter "{p}" type is {type(val)} but should be {param_signature[p].annotation} [{f!r}]')

            return f(*args, **kwargs)
        return wrapper
    if arg_is_func:
        return wrap(func)
    return wrap


def takes_context(context_arg_name='context', optional=True):
    arg_is_func = False
    func = None
    if isinstance(context_arg_name, typing.Callable):
        arg_is_func = True
        func = context_arg_name
        context_arg_name = 'context'

    def wrap(f):
        param_signature: typing.Dict[str, inspect.Parameter] = dict(inspect.signature(f).parameters)

        if context_arg_name not in param_signature:
            if not optional:
                raise NameError(f'Context argument {context_arg_name} (non-optional) not found in function {f!r} -> {param_signature!r}')
        else:
            setattr(f, '__takes_context_as__', context_arg_name)

        return f
    if arg_is_func:
        return wrap(func)
    return wrap
